#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines the tree editor factory for all traits user interface toolkits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Any, Dict, Bool, Tuple, Int, List, Instance, Str, Enum

from ..tree_node import TreeNode

from ..dock_window_theme import DockWindowTheme

from ..editor_factory import EditorFactory

from ..helper import Orientation

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Size of each tree node icon
IconSize = Tuple((16, 16), Int, Int)

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for tree editors.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Supported TreeNode objects
    nodes = List(TreeNode)

    # Mapping from TreeNode tuples to MultiTreeNodes
    multi_nodes = Dict

    # The column header labels if any.
    column_headers = List(Str)

    # Are the individual nodes editable?
    editable = Bool(True)

    # Selection mode.
    selection_mode = Enum('single', 'extended')

    # Is the editor shared across trees?
    shared_editor = Bool(False)

    # Reference to a shared object editor
    editor = Instance(EditorFactory)

    # The DockWindow graphical theme
    # FIXME: Implemented only in wx backend.
    dock_theme = Instance(DockWindowTheme)

    # Show icons for tree nodes?
    show_icons = Bool(True)

    # Hide the tree root node?
    hide_root = Bool(False)

    # Layout orientation of the tree and the editor
    orientation = Orientation

    # Number of tree levels (down from the root) that should be automatically
    # opened
    auto_open = Int

    # Size of the tree node icons
    icon_size = IconSize

    # Called when a node is selected
    on_select = Any

    # Called when a node is clicked
    on_click = Any

    # Called when a node is double-clicked
    on_dclick = Any

    # Called when a node is activated
    on_activated = Any

    # Call when the mouse hovers over a node
    on_hover = Any

    # The optional extended trait name of the trait to synchronize with the
    # editor's current selection:
    selected = Str

    # The optional extended trait name of the trait that should be assigned
    # a node object when a tree node is activated, by double-clicking or
    # pressing the Enter key when a node has focus (Note: if you want to
    # receive repeated activated events on the same node, make sure the trait
    # is defined as an Event):
    activated = Str

    # The optional extended trait name of the trait that should be assigned
    # a node object when a tree node is clicked on (Note: If you want to
    # receive repeated clicks on the same node, make sure the trait is defined
    # as an Event):
    click = Str

    # The optional extended trait name of the trait that should be assigned
    # a node object when a tree node is double-clicked on (Note: if you want to
    # receive repeated double-clicks on the same node, make sure the trait is
    # defined as an Event):
    dclick = Str

    # The optional extended trait name of the trait event that is fired
    # whenever the application wishes to veto a tree action in progress (e.g.
    # double-clicking a non-leaf tree node normally opens or closes the node,
    # but if you are handling the double-click event in your program, you may
    # wish to veto the open or close operation). Be sure to fire the veto event
    # in the event handler triggered by the operation (e.g. the 'dclick' event
    # handler.
    veto = Str

    # The optional extended trait name of the trait event that is fired when the
    # application wishes the currently visible portion of the tree widget to
    # repaint itself.
    refresh = Str

    # Mode for lines connecting tree nodes
    #
    # * 'appearance': Show lines only when they look good.
    # * 'on': Always show lines.
    # * 'off': Don't show lines.
    lines_mode = Enum('appearance', 'on', 'off')
    # FIXME: Document as unimplemented or wx specific.
    # Whether to alternate row colors or not.
    alternating_row_colors = Bool(False)

    # Any extra vertical padding to add.
    vertical_padding = Int(0)

    # Whether or not to expand on a double-click.
    expands_on_dclick = Bool(True)

    # Whether the labels should be wrapped around, if not an ellipsis is shown
    # This works only in the qt backend and if there is only one column in tree
    word_wrap = Bool(False)

# Define the TreeEditor class.
TreeEditor = ToolkitEditorFactory

### EOF #######################################################################



""" Utility functions for help plugin actions.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

from os.path import isabs, join, normpath
import sys

def get_sys_prefix_relative_filename(filename):
    return None if (filename is None) else \
           filename if isabs(filename) else \
           normpath(join(sys.prefix, filename))



"""Simple class to support file path objects that work well in the
context of persistent storage with the state_pickler.

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import os
from os.path import abspath, normpath, dirname, commonprefix, join


class FilePath(object):
    """This class stores two paths to the file.  A relative path and
    an absolute one.  The absolute path is used by the end user.  When
    this object is pickled the state_pickler sets the relative path
    relative to the file that is being generated.  When unpickled, the
    stored relative path is used to set the absolute path correctly
    based on the path of the saved file.
    """
    def __init__(self, value=''):
        self.set(value)

    def __str__(self):
        return self.abs_pth

    def __repr__(self):
        return self.abs_pth.__repr__()

    def get(self):
        """Get the path.
        """
        return self.abs_pth

    def set(self, value):
        """Sets the value of the path.
        """
        self.rel_pth = value
        if value:
            self.abs_pth = normpath(abspath(value))
        else:
            self.abs_pth = ''

    def set_relative(self, base_f_name):
        """Sets the path relative to `base_f_name`.  Note that
        `base_f_name` and self.rel_pth should be valid file names
        correct on the current os.  The set name is a file name that
        has a POSIX path.
        """

        # Get normalized paths.
        _src = abspath(base_f_name)
        _dst = self.abs_pth

        # Now strip out any common prefix between the two paths.
        for part in _src.split(os.sep):
            if _dst.startswith(part+os.sep):
                length = len(part) + 1
                _src = _src[length:]
                _dst = _dst[length:]
            else:
                break

        # For each directory in the source, we need to add a reference to
        # the parent directory to the destination.
        ret = (_src.count(os.sep) * ('..' + os.sep)) + _dst

        # Make it posix style.
        if os.sep is not '/':
            ret.replace(os.sep, '/')

        # Store it.
        self.rel_pth = ret

    def set_absolute(self, base_f_name):
        """Sets the absolute file name for the current relative file
        name with respect to the given `base_f_name`.
        """
        base_f_name = normpath(abspath(base_f_name))
        rel_file_name = normpath(self.rel_pth)
        file_name = join(dirname(base_f_name), rel_file_name)
        file_name = os.path.normpath(file_name)
        self.abs_pth = file_name




#------------------------------------------------------------------------------
# Copyright (c) 2005-2015, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The preference dialog. """


# Major package imports.
import wx

# Enthought library imports.
from traits.api import Any, Dict, Float, Instance, Str

# Local imports.
from pyface.preference.preference_node import PreferenceNode
from pyface.ui.wx.heading_text import HeadingText
from pyface.ui.wx.layered_panel import LayeredPanel
from pyface.ui.wx.split_dialog import SplitDialog
from pyface.ui.wx.viewer.tree_viewer import TreeViewer
from pyface.viewer.default_tree_content_provider import DefaultTreeContentProvider
from pyface.wx.util.font_helper import new_font_like


class PreferenceDialog(SplitDialog):
    """ The preference dialog. """

    #### 'Dialog' interface ###################################################

    # The dialog title.
    title = Str('Preferences')

    #### 'SplitDialog' interface ##############################################

    # The ratio of the size of the left/top pane to the right/bottom pane.
    ratio = Float(0.25)

    #### 'PreferenceDialog' interface #########################################

    # The root of the preference hierarchy.
    root = Instance(PreferenceNode)

    #### Private interface ####################################################

    # The preference pages in the dialog (they are created lazily).
    _pages = Dict

    # The current visible preference page.
    _current_page = Any

    ###########################################################################
    # Protected 'Dialog' interface.
    ###########################################################################

    def _create_buttons(self, parent):
        """ Creates the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Done' button.
        done = wx.Button(parent, wx.ID_OK, "Done")
        done.SetDefault()
        wx.EVT_BUTTON(parent, wx.ID_OK, self._wx_on_ok)
        sizer.Add(done)

        return sizer

    ###########################################################################
    # Protected 'SplitDialog' interface.
    ###########################################################################

    def _create_lhs(self, parent):
        """ Creates the panel containing the preference page tree. """

        return self._create_tree(parent)

    def _create_rhs(self, parent):
        """ Creates the panel containing the selected preference page. """

        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # The 'pretty' title bar ;^)
        self.__title = HeadingText(panel)
        sizer.Add(
            self.__title.control, 0,
            wx.EXPAND | wx.BOTTOM | wx.LEFT | wx.RIGHT, 5
        )

        # The preference page of the node currently selected in the tree.
        self._layered_panel = LayeredPanel(panel, min_width=-1, min_height=-1)
        sizer.Add(
            self._layered_panel.control, 1,
            wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, 5
        )

        # The 'Restore Defaults' button etc.
        buttons = self._create_page_buttons(panel)
        sizer.Add(buttons, 0, wx.ALIGN_RIGHT | wx.TOP | wx.RIGHT, 5)

        # A separator.
        line = wx.StaticLine(panel, -1)
        sizer.Add(line, 0, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, 5)

        # Resize the panel to fit the sizer's minimum size.
        sizer.Fit(panel)

        return panel

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_tree(self, parent):
        """ Creates the preference page tree. """

        tree_viewer = TreeViewer(
            parent,
            input            = self.root,
            show_images      = False,
            show_root        = False,
            content_provider = DefaultTreeContentProvider()
        )

        tree_viewer.on_trait_change(self._on_selection_changed, 'selection')

        return tree_viewer.control

    def _create_page_buttons(self, parent):
        """ Creates the 'Restore Defaults' button, etc.

        At the moment the "etc." is an optional 'Help' button.

        """

        self._button_sizer = sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Help' button. Comes first so 'Restore Defaults' doesn't jump around.
        self._help = help = wx.Button(parent, -1, "Help")
        wx.EVT_BUTTON(parent, help.GetId(), self._on_help)
        sizer.Add(help, 0, wx.RIGHT, 5)

        # 'Restore Defaults' button.
        restore = wx.Button(parent, -1, "Restore Defaults")
        wx.EVT_BUTTON(parent, restore.GetId(), self._on_restore_defaults)
        sizer.Add(restore)

        return sizer

    ###########################################################################
    # wx event handlers.
    ###########################################################################

    def _on_restore_defaults(self, event):
        """ Called when the 'Restore Defaults' button is pressed. """

        page = self._pages[self._layered_panel.current_layer_name]
        page.restore_defaults()

        return

    def _on_help(self, event):
        """ Called when the 'Help' button is pressed. """

        page = self._pages[self._layered_panel.current_layer_name]
        page.show_help_topic()

        return

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _on_selection_changed(self, selection):
        """ Called when a node in the tree is selected. """

        if len(selection) > 0:
            # The tree is in single selection mode.
            node = selection[0]

            # We only show the help button if the selected node has a help
            # topic Id.
            if len(node.help_id) > 0:
                self._button_sizer.Show(self._help, True)

            else:
                self._button_sizer.Show(self._help, False)

            # Show the selected preference page.
            layered_panel  = self._layered_panel
            parent = self._layered_panel.control

            # If we haven't yet displayed the node's preference page during the
            # lifetime of this dialog, then we have to create it.
            if not layered_panel.has_layer(node.name):
                page = node.create_page()
                layered_panel.add_layer(node.name, page.create_control(parent))
                self._pages[node.name] = page

            layered_panel.show_layer(node.name)
            self.__title.text = node.name

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The base class for all pyface wizards. """


# Major package imports.
import wx

# Enthought library imports.
from traits.api import Bool, Instance, List, Property, provides, Unicode
from pyface.api import Dialog, LayeredPanel
from pyface.wizard.i_wizard import IWizard, MWizard
from pyface.wizard.i_wizard_controller import IWizardController
from pyface.wizard.i_wizard_page import IWizardPage


@provides(IWizard)
class Wizard(MWizard, Dialog):
    """ The base class for all pyface wizards.

    See the IWizard interface for the API documentation.

    """


    #### 'IWizard' interface ##################################################

    pages = Property(List(IWizardPage))

    controller = Instance(IWizardController)

    show_cancel = Bool(True)

    #### 'IWindow' interface ##################################################

    title = Unicode('Wizard')

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def _create_dialog_area(self, parent):
        """ Creates the main content of the dialog. """

        self._layered_panel = panel = LayeredPanel(parent)
        # fixme: Specific size?
        panel.control.SetSize((100, 200))

        return panel.control

    def _create_buttons(self, parent):
        """ Creates the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Back' button.
        self._back = back = wx.Button(parent, -1, "Back")
        wx.EVT_BUTTON(parent, back.GetId(), self._on_back)
        sizer.Add(back, 0)

        # 'Next' button.
        self._next = next = wx.Button(parent, -1, "Next")
        wx.EVT_BUTTON(parent, next.GetId(), self._on_next)
        sizer.Add(next, 0, wx.LEFT, 5)
        next.SetDefault()

        # 'Finish' button.
        self._finish = finish = wx.Button(parent, wx.ID_OK, "Finish")
        finish.Enable(self.controller.complete)
        wx.EVT_BUTTON(parent, wx.ID_OK, self._wx_on_ok)
        sizer.Add(finish, 0, wx.LEFT, 5)

        # 'Cancel' button.
        if self.show_cancel:
            self._cancel = cancel = wx.Button(parent, wx.ID_CANCEL, "Cancel")
            wx.EVT_BUTTON(parent, wx.ID_CANCEL, self._wx_on_cancel)
            sizer.Add(cancel, 0, wx.LEFT, 10)

        # 'Help' button.
        if len(self.help_id) > 0:
            help = wx.Button(parent, wx.ID_HELP, "Help")
            wx.EVT_BUTTON(parent, wx.ID_HELP, self._wx_on_help)
            sizer.Add(help, 0, wx.LEFT, 10)

        return sizer

    ###########################################################################
    # Protected 'MWizard' interface.
    ###########################################################################

    def _show_page(self, page):
        """ Show the specified page. """

        panel = self._layered_panel

        # If the page has not yet been shown then create it.
        if not panel.has_layer(page.id):
            panel.add_layer(page.id, page.create_page(panel.control))

        # Show the page.
        layer = panel.show_layer(page.id)
        layer.SetFocus()

        # Set the current page in the controller.
        #
        # fixme: Shouldn't this interface be reversed?  Maybe calling
        # 'next_page' on the controller should cause it to set its own current
        # page?
        self.controller.current_page = page

        return

    def _update(self):
        """ Enables/disables buttons depending on the state of the wizard. """

        controller = self.controller
        current_page = controller.current_page

        is_first_page = controller.is_first_page(current_page)
        is_last_page  = controller.is_last_page(current_page)

        # 'Next button'.
        if self._next is not None:
            self._next.Enable(current_page.complete and not is_last_page)

        # 'Back' button.
        if self._back is not None:
            self._back.Enable(not is_first_page)

        # 'Finish' button.
        if self._finish is not None:
            self._finish.Enable(controller.complete)

        # If this is the last page then the 'Finish' button is the default
        # button, otherwise the 'Next' button is the default button.
        if is_last_page:
            if self._finish is not None:
                self._finish.SetDefault()

        else:
            if self._next is not None:
                self._next.SetDefault()

        return

    #### Trait handlers #######################################################

    def _controller_default(self):
        """ Provide a default controller. """

        from pyface.wizard.wizard_controller import WizardController

        return WizardController()

    def _get_pages(self):
        """ Returns the pages in the wizard. """

        return self.controller.pages

    def _set_pages(self, pages):
        """ Sets the pages in the wizard. """

        self.controller.pages = pages

    #### wx event handlers ####################################################

    def _on_next(self, event):
        """ Called when the 'Next' button is pressed. """

        self.next()

        return

    def _on_back(self, event):
        """ Called when the 'Back' button is pressed. """

        self.previous()

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines the tuple editor factory for all traits user interface toolkits.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.trait_base import SequenceTypes

from traits.api import Bool, HasTraits, List, Tuple, Unicode, Int, Any, TraitType

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..group import Group

from ..item import Item

from ..editor_factory import EditorFactory

from ..editor import Editor

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for tuple editors.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Trait definitions for each tuple field
    types = Any

    # Labels for each of the tuple fields
    labels = List(Unicode)

    # Editors for each of the tuple fields:
    editors = List(EditorFactory)

    # Number of tuple fields or rows
    cols = Int(1)

    # Is user input set on every keystroke? This is applied to every field
    # of the tuple, provided the field does not already have an 'auto_set'
    # metadata or an editor defined.
    auto_set = Bool(True)

    # Is user input set when the Enter key is pressed? This is applied to
    # every field of the tuple, provided the field does not already have an
    # 'enter_set' metadata or an editor defined.
    enter_set = Bool(False)

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style of editor for tuples.

    The editor displays an editor for each of the fields in the tuple, based on
    the type of each field.
    """
    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._ts = ts = TupleStructure(self)
        self._ui = ui = ts.view.ui(ts, parent, kind='subpanel').set(
            parent=self.ui)
        self.control = ui.control
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        ts = self._ts
        for i, value in enumerate(self.value):
            setattr(ts, 'f%d' % i, value)

    #-------------------------------------------------------------------------
    #  Returns the editor's control for indicating error status:
    #-------------------------------------------------------------------------

    def get_error_control(self):
        """ Returns the editor's control for indicating error status.
        """
        return self._ui.get_error_controls()

#-------------------------------------------------------------------------
#  'TupleStructure' class:
#-------------------------------------------------------------------------


class TupleStructure(HasTraits):
    """ Creates a view containing items for each field in a tuple.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Editor this structure is linked to
    editor = Any

    # The constructed View for the tuple
    view = Any

    # Number of tuple fields
    fields = Int

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, editor):
        """ Initializes the object.
        """
        factory = editor.factory
        types = factory.types
        labels = factory.labels
        editors = factory.editors
        cols = factory.cols

        # Save the reference to the editor:
        self.editor = editor

        # Get the tuple we are mirroring:
        object = editor.value

        # For each tuple field, add a trait with the appropriate trait
        # definition and default value:
        content = []
        self.fields = len(object)
        len_labels = len(labels)
        len_editors = len(editors)

        if types is None:
            type = editor.value_trait.handler
            if isinstance(type, Tuple):
                types = type.types

        if not isinstance(types, SequenceTypes):
            types = [types]

        len_types = len(types)
        if len_types == 0:
            types = [Any]
            len_types = 1

        for i, value in enumerate(object):
            type = types[i % len_types]

            auto_set = enter_set = None
            if isinstance(type, TraitType):
                auto_set = type.auto_set
                enter_set = type.enter_set
            if auto_set is None:
                auto_set = editor.factory.auto_set
            if enter_set is None:
                enter_set = editor.factory.enter_set

            label = ''
            if i < len_labels:
                label = labels[i]

            field_editor = None
            if i < len_editors:
                field_editor = editors[i]

            name = 'f%d' % i
            self.add_trait(name, type(value, event='field',
                                      auto_set=auto_set,
                                      enter_set=enter_set))
            item = Item(name=name, label=label, editor=field_editor)
            if cols <= 1:
                content.append(item)
            else:
                if (i % cols) == 0:
                    group = Group(orientation='horizontal')
                    content.append(group)

                group.content.append(item)

        self.view = View(Group(show_labels=(len_labels != 0), *content))

    #-------------------------------------------------------------------------
    #  Updates the underlying tuple when any field changes value:
    #-------------------------------------------------------------------------

    def _field_changed(self, name, old, new):
        """ Updates the underlying tuple when any field changes value.
        """
        index = int(name[1:])
        value = self.editor.value
        if new != value[index]:
            self.editor.value = tuple(
                [getattr(self, 'f%d' % i) for i in range(self.fields)])


# Define the TupleEditor class.
TupleEditor = ToolkitEditorFactory

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.action.api import Action
from traits.api import Bool, Event, HasTraits, provides, \
        Instance, List, Unicode

# Local imports.
from apptools.permissions.i_user import IUser
from apptools.permissions.i_user_manager import IUserManager
from apptools.permissions.package_globals import get_permissions_manager
from apptools.permissions.permission import ManageUsersPermission
from i_user_database import IUserDatabase


@provides(IUserManager)
class UserManager(HasTraits):
    """The default user manager implementation."""



    #### 'IUserManager' interface #############################################

    management_actions = List(Instance(Action))

    user = Instance(IUser)

    user_actions = List(Instance(Action))

    user_authenticated = Event(IUser)

    #### 'UserManager' interface ##############################################

    # The user database.
    user_db = Instance(IUserDatabase)

    ###########################################################################
    # 'IUserManager' interface.
    ###########################################################################

    def bootstrapping(self):
        """Return True if we are bootstrapping, ie. no users have been defined.
        """

        return self.user_db.bootstrapping()

    def authenticate_user(self):
        """Authenticate the user."""

        if self.user_db.authenticate_user(self.user):
            self.user.authenticated = True

            # Tell the policy manager before everybody else.
            get_permissions_manager().policy_manager.load_policy(self.user)

            self.user_authenticated = self.user

    def unauthenticate_user(self):
        """Unauthenticate the user."""

        if self.user.authenticated and self.user_db.unauthenticate_user(self.user):
            self.user.authenticated = False

            # Tell the policy manager before everybody else.
            get_permissions_manager().policy_manager.load_policy(None)

            self.user_authenticated = None

    def matching_user(self, name):
        """Select a user."""

        return self.user_db.matching_user(name)

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _management_actions_default(self):
        """Return the list of management actions."""

        from apptools.permissions.secure_proxy import SecureProxy

        user_db = self.user_db
        actions = []
        perm = ManageUsersPermission()

        if user_db.can_add_user:
            act = Action(name="&Add a User...", on_perform=user_db.add_user)
            actions.append(SecureProxy(act, permissions=[perm], show=False))

        if user_db.can_modify_user:
            act = Action(name="&Modify a User...",
                    on_perform=user_db.modify_user)
            actions.append(SecureProxy(act, permissions=[perm], show=False))

        if user_db.can_delete_user:
            act = Action(name="&Delete a User...",
                    on_perform=user_db.delete_user)
            actions.append(SecureProxy(act, permissions=[perm], show=False))

        return actions

    def _user_actions_default(self):
        """Return the list of user actions."""

        actions = []

        if self.user_db.can_change_password:
            actions.append(_ChangePasswordAction())

        return actions

    def _user_default(self):
        """Return the default current user."""

        return self.user_db.user_factory()

    def _user_db_default(self):
        """Return the default user database."""

        # Defer to an external user database if there is one.
        try:
            from apptools.permissions.external.user_database import UserDatabase
        except ImportError:
            from apptools.permissions.default.user_database import UserDatabase

        return UserDatabase()


class _ChangePasswordAction(Action):
    """An action that allows the current user to change their password.  It
    isn't exported through actions/api.py because it is specific to this user
    manager implementation."""

    #### 'Action' interface ###################################################

    enabled = Bool(False)

    name = Unicode("&Change Password...")

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        super(_ChangePasswordAction, self).__init__(**traits)

        get_permissions_manager().user_manager.on_trait_event(self._refresh_enabled, 'user_authenticated')

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """Perform the action."""

        um = get_permissions_manager().user_manager
        um.user_db.change_password(um.user)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _refresh_enabled(self, user):
        """Invoked whenever the current user's authorisation state changes."""

        self.enabled = user is not None



#-------------------------------------------------------------------------------
#
#  A concrete implementation of the ITemplateDataNameItem interface that looks
#  for a specified sub-context in its input context and outputs that as its
#  output context if it is found.
#
#  Written by: David C. Morrill
#
#  Date: 07/29/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A concrete implementation of the ITemplateDataNameItem interface that looks
    for a specified sub-context in its input context and outputs that as its
    output context if it is found.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from apptools.template.template_traits \
    import TStr

from any_context_data_name_item \
    import AnyContextDataNameItem

from helper \
    import path_for, parse_name

#-------------------------------------------------------------------------------
#  'ContextDataNameItem' class:
#-------------------------------------------------------------------------------

class ContextDataNameItem ( AnyContextDataNameItem ):
    """ A concrete implementation of the ITemplateDataNameItem interface that
        looks for a specified sub-context in its input context and outputs that
        as its output context if it is found.
    """

    #-- Public Traits ----------------------------------------------------------

    # The name of the context to be matched:
    name = TStr

    #-- Abstract Method Implementations ----------------------------------------

    def filter ( self, name, context ):
        """ Returns **True** if the specified *context* called *name* should be
            included in the output context; and **False** otherwise.
        """
        return (name == self.name_last)

    #-- AnyDataNameItem Property Implementation Overrides ----------------------

    def _get_data_name_item_choice ( self ):
        return TemplateChoice( choice_value = self.name )

    def _set_data_name_item_choice ( self, value ):
        self.name = value.choice_value

    def _get_data_name_item_choices ( self ):
        return self._get_choices( self.input_data_context )

    def _get_current_input_data_context ( self ):
        context = self.input_data_context
        for name in parse_name( self.name )[:-1]:
            if name not in context.data_contexts:
                return None

            context = context.get_data_context( name )

        return context

    #-- Trait Event Handlers ---------------------------------------------------

    def _name_changed ( self, name ):
        """ Handles the 'name' trait being changed.
        """
        self.name_last = parse_name( name )[-1]
        self.inputs_changed()

    #-- Private Methods --------------------------------------------------------

    def _get_choices ( self, context, path = '' ):
        """ Returns all of the valid TemplateChoices for this item.
        """
        choices = []
        gdc     = context.get_data_context
        for name in context.data_contexts:
            next_path = path_for( path, name )
            choices.append( TemplateChoice( choice_value = next_path ) )
            choices.extend( self._get_choices( gdc( name ), next_path ) )

        return choices




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Utility functions for working with wx Fonts. """


# Major package imports.
import wx


def new_font_like(font, **kw):
    """ Creates a new font, like another one, only different.  Maybe. """

    point_size = kw.get('point_size', font.GetPointSize())
    family     = kw.get('family', font.GetFamily())
    style      = kw.get('style', font.GetStyle())
    weight     = kw.get('weight', font.GetWeight())
    underline  = kw.get('underline', font.GetUnderlined())
    face_name  = kw.get('face_name', font.GetFaceName())

    return wx.Font(point_size, family, style, weight, underline, face_name)

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A widget for editing Python code. """


# Enthought library imports.
from traits.api import Bool, Event, Instance, File, Interface, Unicode
from pyface.tasks.i_editor import IEditor

# Local imports.
from pyface.key_pressed_event import KeyPressedEvent


class IStyledTextEditor(IEditor):
    """ A widget for editing text. """

    #### 'IPythonEditor' interface ############################################

    # Object being editor is a file
    obj = Instance(File)

    # The pathname of the file being edited.
    path = Unicode

    # Should line numbers be shown in the margin?
    show_line_numbers = Bool(True)

    #### Events ####

    # The contents of the editor has changed.
    changed = Event

    # A key has been pressed.
    key_pressed = Event(KeyPressedEvent)

    ###########################################################################
    # 'IPythonEditor' interface.
    ###########################################################################

    def load(self, path=None):
        """ Loads the contents of the editor. """

    def save(self, path=None):
        """ Saves the contents of the editor. """

    def select_line(self, lineno):
        """ Selects the specified line. """



# Enthought library imports.
from traits.api import Callable, HasTraits, Str, Enum


class SchemaAddition(HasTraits):
    """ An addition to an existing menu bar or tool bar schema.
    """

    # The schema addition's identifier. This optional, but if left unspecified,
    # other schema additions will be unable to refer to this one.
    id = Str

    # A callable to create the item. Should have the following signature:
    #    callable() -> Action | ActionItem | Group | MenuManager | 
    #                  GroupSchema | MenuSchema
    # If the result is a schema, it will itself admit of extension by other
    # additions. If not, the result will be fixed.
    factory = Callable

    # A forward-slash-separated path through the action hierarchy to the menu
    # to add the action, group or menu to. For example:
    # - To add an item to the menu bar: ``path = "MenuBar"``
    # - To add an item to the tool bar: ``path = "ToolBar"``
    # - To add an item to a sub-menu: ``path = "MenuBar/File/New"``
    path = Str

    # The item appears after the item with this ID.
    # - for groups, this is the ID of another group.
    # - for menus and actions, this is the ID of another menu or action.
    after = Str

    # The action appears before the item with this ID.
    # - for groups, this is the ID of another group.
    # - for menus and actions, this is the ID of another menu or action.
    before = Str

    # The action appears at the absolute specified position first or last.
    # This is useful for example to keep the File menu the first menu in a
    # menubar, the help menu the last etc.
    # If multiple actions in a schema have absolute_position 'first', they
    # will appear in the same order specified; unless 'before' and 'after'
    # traits are set to sort these multiple items.
    # This trait takes precedence over 'after' and 'before', and values of
    # those traits that are not compatible with the  absolute_position are
    # ignored.
    absolute_position = Enum(None, 'first', 'last')



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Bool, HasTraits, Property, Str, Unicode

# Locals imports.
from package_globals import get_permissions_manager


class Permission(HasTraits):
    """A permission is the link between an application action and the current
    user - if the user has a permission attached to the action then the user is
    allowed to perform that action."""

    #### 'Permission' interface ###############################################

    # The id of the permission.  By convention a dotted format is used for the
    # id with the id of the application being the first part.
    id = Str

    # A user friendly description of the permission.
    description = Unicode

    # Set if the current user has this permission.  This is typically used with
    # the enabled_when and visible_when traits of a TraitsUI Item object when
    # the permission instance has been placed in the TraitsUI context.
    granted = Property

    # Set if the permission should be granted automatically when bootstrapping.
    # This is normally only ever set for permissions related to user management
    # and permissions.  The user manager determines exactly what is meant by
    # "bootstrapping" but it is usually when it determines that no user or
    # permissions information has been defined.
    bootstrap = Bool(False)

    # Set if the permission has been defined by application code rather than as
    # a result of loading the policy database.
    application_defined = Bool(True)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """Initialise the object."""

        super(Permission, self).__init__(**traits)

        # Register the permission.
        get_permissions_manager().policy_manager.register_permission(self)

    def __str__(self):
        """Return a user friendly representation."""

        s = self.description
        if not s:
            s = self.id

        return s

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _get_granted(self):
        """Check the user has this permission."""

        return get_permissions_manager().check_permissions(self)


class ManagePolicyPermission(Permission):
    """The standard permission for managing permissions policies."""

    #### 'Permission' interface ###############################################

    id = Str('ets.permissions.manage_policy')

    description = Unicode(u"Manage permissions policy")

    bootstrap = Bool(True)


class ManageUsersPermission(Permission):
    """The standard permission for managing permissions users."""

    #### 'Permission' interface ###############################################

    id = Str('ets.permissions.manage_users')

    description = Unicode(u"Manage users")

    bootstrap = Bool(True)



# Import the toolkit specific version.
import os
import sys

_toolkit_backend_name = None


def _init_toolkit():
    from pyface.toolkit import _toolkit_backend
    backend = _toolkit_backend.strip(".").split(".")[-1]

    global _toolkit_backend_name
    _toolkit_backend_name = "_" + backend


# Do this once then disappear.
_init_toolkit()
del _init_toolkit


def toolkit_object(mname, oname):
    """Add the toolkit specific object with the given name to the namespace given by mname.
    
    This paradigm is different than the Enthough/Pyface library that puts
    the toolkit-specific code in ui/[wx|qt4]/ subdirectories.  This function
    imports toolkit-specific code from files adjacent to the generic code.
    For example, if the generic module is named test_widget.py, the toolkit-
    specific code should reside in test_widget_wx.py and test_widget_qt4.py in
    the same directory.
    
    Typical use of this function would be to use the following two-line block
    of code in the generic module (e.g.  test_widget.py):
    
        from framework_toolkit import toolkit_object
        toolkit_object(__name__, 'TestWidget')
    
    :param str mname: the full module name of the generic module, typically __name__
    :param str oname: The object name to import as the toolkit-specific version of the generic module
    """

    be_mname = mname + _toolkit_backend_name

    class Unimplemented(object):
        """ This is returned if an object isn't implemented by the selected
        toolkit.  It raises an exception if it is ever instantiated.
        """

        def __init__(self, *args, **kwargs):
            raise NotImplementedError("the %s pyface backend doesn't implement %s" % (ETSConfig.toolkit, oname))

    be_obj = Unimplemented

    try:
        __import__(be_mname)

        try:
            be_obj = getattr(sys.modules[be_mname], oname)
            setattr(sys.modules[mname], oname, be_obj)
        except AttributeError:
            pass
    except ImportError, e:

        # Ignore *ANY* errors unless a debug ENV variable is set.
        if 'ETS_DEBUG' in os.environ:
            raise



#-------------------------------------------------------------------------------
#
#  Publically exported symbols for the template.impl package.
#
#  Written by: David C. Morrill
#
#  Date: 07/29/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

from any_context_data_name_item \
    import AnyContextDataNameItem

from any_data_name_item \
    import AnyDataNameItem

from context_data_name_item \
    import ContextDataNameItem

from template_data_context \
    import TemplateDataContext

from template_data_source \
    import TemplateDataSource

from value_data_name_item \
    import ValueDataNameItem

from value_nd_data_name_item \
    import ValueNDDataNameItem, Value1DDataNameItem, Value2DDataNameItem, \
           Value3DDataNameItem




# Enthought library imports.
from pyface.action.api import Action, ActionItem, Group
from traits.api import Any, Instance, List, Property, Unicode, \
     on_trait_change

# Local imports.
from envisage._compat import unicode_str


class TaskWindowToggleAction(Action):
    """ An action for activating an application window.
    """

    #### 'Action' interface ###################################################

    name = Property(Unicode, depends_on='window.active_task.name')
    style = 'toggle'

    #### 'TaskWindowToggleAction' interface ###################################

    # The window to use for this action.
    window = Instance('envisage.ui.tasks.task_window.TaskWindow')

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event=None):
        if self.window:
            self.window.activate()

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_name(self):
        if self.window.active_task:
            return self.window.active_task.name
        return unicode_str()

    @on_trait_change('window:activated')
    def _window_activated(self):
        self.checked = True

    @on_trait_change('window:deactivated')
    def _window_deactivated(self):
        self.checked = False


class TaskWindowToggleGroup(Group):
    """ A Group for toggling the activation state of an application's windows.
    """

    #### 'Group' interface ####################################################

    id = 'TaskWindowToggleGroup'
    items = List

    #### 'TaskWindowToggleGroup' interface ####################################

    # The application that contains the group.
    application = Instance('envisage.ui.tasks.tasks_application.'
                           'TasksApplication')

    # The ActionManager to which the group belongs.
    manager = Any

    ###########################################################################
    # 'Group' interface.
    ###########################################################################

    def destroy(self):
        """ Called when the group is no longer required.
        """
        super(TaskWindowToggleGroup, self).destroy()
        if self.application:
            self.application.on_trait_change(
                self._rebuild, 'window_opened, window_closed', remove=True)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self):
        items = []
        for window in self.application.windows:
            active = window == self.application.active_window
            action = TaskWindowToggleAction(window=window, checked=active)
            items.append(ActionItem(action=action))
        return items

    def _rebuild(self):
        # Clear out the old group, then build the new one.
        for item in self.items:
            item.destroy()
        self.items = self._get_items()

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    #### Trait initializers ###################################################

    def _application_default(self):
        return self.manager.controller.task.window.application

    def _items_default(self):
        self.application.on_trait_change(self._rebuild,
                                         'window_opened, window_closed')
        return self._get_items()

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        return manager



import re
import svg_constants

re_list_parser = re.compile("([A-Za-z]|-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)")

re_func_parser = re.compile('\w+\((?:\-?[0-9]+(?:\.[0-9]*)?\w*\s*)?(?:\s*,\s*\-?\s*[0-9]*(?:\.[0-9]+)?\w*\s*)*\)')

def get_fns(string):
    string = string.strip()
    return re_func_parser.findall(string)

def parse_list(string):
    return re_list_parser.findall(string)

def parse_float_list(string):
    return [parse_float(x.strip()) for x in string.split(',')]

def parse_style(string):
    s_dict = {}
    for item in string.split(';'):
        if ':' in item:
            key, value = item.split(':')
            s_dict[key.strip()] = value.strip()
    return s_dict

def parse_float(txt):
    #assume 90 dpi
    if txt.endswith('%'):
        pct = float(txt[:-1])/100.0
        return pct
    elif txt.endswith('px'):
        return float(txt[:-2])
    elif txt.endswith('pt'):
        return float(txt[:-2]) * 1.25
    elif txt.endswith('pc'):
        return float(txt[:-2]) * 15
    elif txt.endswith('mm'):
        return float(txt[:-2]) * 3.543307
    elif txt.endswith('cm'):
        return float(txt[:-2]) * 35.43307
    elif txt.endswith('in'):
        return float(txt[:-2]) * 90
    elif txt.endswith('em') or txt.endswith('ex'):
        return float(txt[:-2]) * 10
    else:
        return float(txt)

def parse_transform(txt):
    pass

def parse_color(c, default=None):
    if not c:
        return default
    if c == 'none':
        return None

    c = c.strip()

    if c in svg_constants.named_colors:
        c = svg_constants.named_colors[c]

    if c.startswith('rgb'):
        start = c.index('(')
        end = c.index(')')
        parts = c[start+1:end].split(',')
        r, g, b = tuple(int(p.strip()) for p in parts)
        return [r,g,b,255]

    if c[0] == '#': c = c[1:]
    if c.startswith('url(#'):
        return c[5:-1]
    try:
        a=255
        if len(c) == 8:
            r = int(c[0:2], 16)
            g = int(c[2:4], 16)
            b = int(c[4:6], 16)
            a = int(c[6:8], 16)
        elif len(c) == 6:
            r = int(c[0:2], 16)
            g = int(c[2:4], 16)
            b = int(c[4:6], 16)
        elif len(c) == 3:
            r = int(c[0], 16) * 17
            g = int(c[1], 16) * 17
            b = int(c[2], 16) * 17
        else:
            raise Exception("Incorrect length for color " + str(c) + " length " + str(len(c)))
        return [r,g,b,a]
    except Exception, ex:
        print 'Exception parsing color', ex
        return None
        


#!/usr/bin/env python
from fs.commands.runner import Command
import sys

class FSCat(Command):
    
    usage = """fscat [OPTION]... [FILE]...
Concetanate FILE(s)"""

    version = "1.0"
    
    def do_run(self, options, args):
        count = 0            
        for fs, path, is_dir in self.get_resources(args):            
            if is_dir:
                self.error('%s is a directory\n' % path)
                return 1                      
            self.output(fs.getcontents(path))
            count += 1              
        if self.is_terminal() and count:
            self.output('\n')
    
def run():
    return FSCat().run()
    
if __name__ == "__main__":
    sys.exit(run())
        


# Copyright (c) 2008 Giorgos Verigakis <verigak@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from ctypes import *
from ctypes.util import find_library
from errno import *
from functools import partial
from platform import machine, system
from stat import S_IFDIR
from traceback import print_exc

import logging


class c_timespec(Structure):
    _fields_ = [('tv_sec', c_long), ('tv_nsec', c_long)]

class c_utimbuf(Structure):
    _fields_ = [('actime', c_timespec), ('modtime', c_timespec)]

class c_stat(Structure):
    pass    # Platform dependent

_system = system()
if _system in ('Darwin', 'FreeBSD'):
    _libiconv = CDLL(find_library("iconv"), RTLD_GLOBAL)     # libfuse dependency
    ENOTSUP = 45
    c_dev_t = c_int32
    c_fsblkcnt_t = c_ulong
    c_fsfilcnt_t = c_ulong
    c_gid_t = c_uint32
    c_mode_t = c_uint16
    c_off_t = c_int64
    c_pid_t = c_int32
    c_uid_t = c_uint32
    setxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte),
        c_size_t, c_int, c_uint32)
    getxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte),
        c_size_t, c_uint32)
    c_stat._fields_ = [
        ('st_dev', c_dev_t),
        ('st_ino', c_uint32),
        ('st_mode', c_mode_t),
        ('st_nlink', c_uint16),
        ('st_uid', c_uid_t),
        ('st_gid', c_gid_t),
        ('st_rdev', c_dev_t),
        ('st_atimespec', c_timespec),
        ('st_mtimespec', c_timespec),
        ('st_ctimespec', c_timespec),
        ('st_size', c_off_t),
        ('st_blocks', c_int64),
        ('st_blksize', c_int32)]
elif _system == 'Linux':
    ENOTSUP = 95
    c_dev_t = c_ulonglong
    c_fsblkcnt_t = c_ulonglong
    c_fsfilcnt_t = c_ulonglong
    c_gid_t = c_uint
    c_mode_t = c_uint
    c_off_t = c_longlong
    c_pid_t = c_int
    c_uid_t = c_uint
    setxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t, c_int)
    getxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t)

    _machine = machine()
    if _machine == 'x86_64':
        c_stat._fields_ = [
            ('st_dev', c_dev_t),
            ('st_ino', c_ulong),
            ('st_nlink', c_ulong),
            ('st_mode', c_mode_t),
            ('st_uid', c_uid_t),
            ('st_gid', c_gid_t),
            ('__pad0', c_int),
            ('st_rdev', c_dev_t),
            ('st_size', c_off_t),
            ('st_blksize', c_long),
            ('st_blocks', c_long),
            ('st_atimespec', c_timespec),
            ('st_mtimespec', c_timespec),
            ('st_ctimespec', c_timespec)]
    elif _machine == 'ppc':
        c_stat._fields_ = [
            ('st_dev', c_dev_t),
            ('st_ino', c_ulonglong),
            ('st_mode', c_mode_t),
            ('st_nlink', c_uint),
            ('st_uid', c_uid_t),
            ('st_gid', c_gid_t),
            ('st_rdev', c_dev_t),
            ('__pad2', c_ushort),
            ('st_size', c_off_t),
            ('st_blksize', c_long),
            ('st_blocks', c_longlong),
            ('st_atimespec', c_timespec),
            ('st_mtimespec', c_timespec),
            ('st_ctimespec', c_timespec)]
    else:
        # i686, use as fallback for everything else
        c_stat._fields_ = [
            ('st_dev', c_dev_t),
            ('__pad1', c_ushort),
            ('__st_ino', c_ulong),
            ('st_mode', c_mode_t),
            ('st_nlink', c_uint),
            ('st_uid', c_uid_t),
            ('st_gid', c_gid_t),
            ('st_rdev', c_dev_t),
            ('__pad2', c_ushort),
            ('st_size', c_off_t),
            ('st_blksize', c_long),
            ('st_blocks', c_longlong),
            ('st_atimespec', c_timespec),
            ('st_mtimespec', c_timespec),
            ('st_ctimespec', c_timespec),
            ('st_ino', c_ulonglong)]
else:
    raise NotImplementedError('%s is not supported.' % _system)


class c_statvfs(Structure):
    _fields_ = [
        ('f_bsize', c_ulong),
        ('f_frsize', c_ulong),
        ('f_blocks', c_fsblkcnt_t),
        ('f_bfree', c_fsblkcnt_t),
        ('f_bavail', c_fsblkcnt_t),
        ('f_files', c_fsfilcnt_t),
        ('f_ffree', c_fsfilcnt_t),
        ('f_favail', c_fsfilcnt_t)]

if _system == 'FreeBSD':
    c_fsblkcnt_t = c_uint64
    c_fsfilcnt_t = c_uint64
    setxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t, c_int)
    getxattr_t = CFUNCTYPE(c_int, c_char_p, c_char_p, POINTER(c_byte), c_size_t)
    class c_statvfs(Structure):
        _fields_ = [
            ('f_bavail', c_fsblkcnt_t),
            ('f_bfree', c_fsblkcnt_t),
            ('f_blocks', c_fsblkcnt_t),
            ('f_favail', c_fsfilcnt_t),
            ('f_ffree', c_fsfilcnt_t),
            ('f_files', c_fsfilcnt_t),
            ('f_bsize', c_ulong),
            ('f_flag', c_ulong),
            ('f_frsize', c_ulong)]

class fuse_file_info(Structure):
    _fields_ = [
        ('flags', c_int),
        ('fh_old', c_ulong),
        ('writepage', c_int),
        ('direct_io', c_uint, 1),
        ('keep_cache', c_uint, 1),
        ('flush', c_uint, 1),
        ('padding', c_uint, 29),
        ('fh', c_uint64),
        ('lock_owner', c_uint64)]

class fuse_context(Structure):
    _fields_ = [
        ('fuse', c_voidp),
        ('uid', c_uid_t),
        ('gid', c_gid_t),
        ('pid', c_pid_t),
        ('private_data', c_voidp)]

class fuse_operations(Structure):
    _fields_ = [
        ('getattr', CFUNCTYPE(c_int, c_char_p, POINTER(c_stat))),
        ('readlink', CFUNCTYPE(c_int, c_char_p, POINTER(c_byte), c_size_t)),
        ('getdir', c_voidp),    # Deprecated, use readdir
        ('mknod', CFUNCTYPE(c_int, c_char_p, c_mode_t, c_dev_t)),
        ('mkdir', CFUNCTYPE(c_int, c_char_p, c_mode_t)),
        ('unlink', CFUNCTYPE(c_int, c_char_p)),
        ('rmdir', CFUNCTYPE(c_int, c_char_p)),
        ('symlink', CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('rename', CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('link', CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('chmod', CFUNCTYPE(c_int, c_char_p, c_mode_t)),
        ('chown', CFUNCTYPE(c_int, c_char_p, c_uid_t, c_gid_t)),
        ('truncate', CFUNCTYPE(c_int, c_char_p, c_off_t)),
        ('utime', c_voidp),     # Deprecated, use utimens
        ('open', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info))),
        ('read', CFUNCTYPE(c_int, c_char_p, POINTER(c_byte), c_size_t, c_off_t,
            POINTER(fuse_file_info))),
        ('write', CFUNCTYPE(c_int, c_char_p, POINTER(c_byte), c_size_t, c_off_t,
            POINTER(fuse_file_info))),
        ('statfs', CFUNCTYPE(c_int, c_char_p, POINTER(c_statvfs))),
        ('flush', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info))),
        ('release', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info))),
        ('fsync', CFUNCTYPE(c_int, c_char_p, c_int, POINTER(fuse_file_info))),
        ('setxattr', setxattr_t),
        ('getxattr', getxattr_t),
        ('listxattr', CFUNCTYPE(c_int, c_char_p, POINTER(c_byte), c_size_t)),
        ('removexattr', CFUNCTYPE(c_int, c_char_p, c_char_p)),
        ('opendir', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info))),
        ('readdir', CFUNCTYPE(c_int, c_char_p, c_voidp, CFUNCTYPE(c_int, c_voidp,
            c_char_p, POINTER(c_stat), c_off_t), c_off_t, POINTER(fuse_file_info))),
        ('releasedir', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info))),
        ('fsyncdir', CFUNCTYPE(c_int, c_char_p, c_int, POINTER(fuse_file_info))),
        ('init', CFUNCTYPE(c_voidp, c_voidp)),
        ('destroy', CFUNCTYPE(c_voidp, c_voidp)),
        ('access', CFUNCTYPE(c_int, c_char_p, c_int)),
        ('create', CFUNCTYPE(c_int, c_char_p, c_mode_t, POINTER(fuse_file_info))),
        ('ftruncate', CFUNCTYPE(c_int, c_char_p, c_off_t, POINTER(fuse_file_info))),
        ('fgetattr', CFUNCTYPE(c_int, c_char_p, POINTER(c_stat),
            POINTER(fuse_file_info))),
        ('lock', CFUNCTYPE(c_int, c_char_p, POINTER(fuse_file_info), c_int, c_voidp)),
        ('utimens', CFUNCTYPE(c_int, c_char_p, POINTER(c_utimbuf))),
        ('bmap', CFUNCTYPE(c_int, c_char_p, c_size_t, POINTER(c_ulonglong)))]


def time_of_timespec(ts):
    return ts.tv_sec + ts.tv_nsec / 10 ** 9

def set_st_attrs(st, attrs):
    for key, val in attrs.items():
        if key in ('st_atime', 'st_mtime', 'st_ctime'):
            timespec = getattr(st, key + 'spec')
            timespec.tv_sec = int(val)
            timespec.tv_nsec = int((val - timespec.tv_sec) * 10 ** 9)
        elif hasattr(st, key):
            setattr(st, key, val)


_libfuse_path = find_library('fuse')
if not _libfuse_path:
    raise EnvironmentError('Unable to find libfuse')
_libfuse = CDLL(_libfuse_path)
_libfuse.fuse_get_context.restype = POINTER(fuse_context)


def fuse_get_context():
    """Returns a (uid, gid, pid) tuple"""
    ctxp = _libfuse.fuse_get_context()
    ctx = ctxp.contents
    return ctx.uid, ctx.gid, ctx.pid


class FUSE(object):
    """This class is the lower level interface and should not be subclassed
       under normal use. Its methods are called by fuse.
       Assumes API version 2.6 or later."""

    def __init__(self, operations, mountpoint, raw_fi=False, **kwargs):
        """Setting raw_fi to True will cause FUSE to pass the fuse_file_info
           class as is to Operations, instead of just the fh field.
           This gives you access to direct_io, keep_cache, etc."""

        self.operations = operations
        self.raw_fi = raw_fi
        args = ['fuse']
        if kwargs.pop('foreground', False):
            args.append('-f')
        if kwargs.pop('debug', False):
            args.append('-d')
        if kwargs.pop('nothreads', False):
            args.append('-s')
        kwargs.setdefault('fsname', operations.__class__.__name__)
        args.append('-o')
        args.append(','.join(key if val == True else '%s=%s' % (key, val)
            for key, val in kwargs.items()))
        args.append(mountpoint)
        argv = (c_char_p * len(args))(*args)

        fuse_ops = fuse_operations()
        for name, prototype in fuse_operations._fields_:
            if prototype != c_voidp and getattr(operations, name, None):
                op = partial(self._wrapper_, getattr(self, name))
                setattr(fuse_ops, name, prototype(op))
        _libfuse.fuse_main_real(len(args), argv, pointer(fuse_ops),
            sizeof(fuse_ops), None)
        del self.operations     # Invoke the destructor

    def _wrapper_(self, func, *args, **kwargs):
        """Decorator for the methods that follow"""
        try:
            return func(*args, **kwargs) or 0
        except OSError as e:
            return -(e.errno or EFAULT)
        except:
            print_exc()
            return -EFAULT

    def getattr(self, path, buf):
        return self.fgetattr(path, buf, None)

    def readlink(self, path, buf, bufsize):
        ret = self.operations('readlink', path).encode('utf-8')
        data = create_string_buffer(ret[:bufsize - 1])
        memmove(buf, data, len(data))
        return 0

    def mknod(self, path, mode, dev):
        return self.operations('mknod', path, mode, dev)

    def mkdir(self, path, mode):
        return self.operations('mkdir', path, mode)

    def unlink(self, path):
        return self.operations('unlink', path)

    def rmdir(self, path):
        return self.operations('rmdir', path)

    def symlink(self, source, target):
        return self.operations('symlink', target, source)

    def rename(self, old, new):
        return self.operations('rename', old, new)

    def link(self, source, target):
        return self.operations('link', target, source)

    def chmod(self, path, mode):
        return self.operations('chmod', path, mode)

    def chown(self, path, uid, gid):
        return self.operations('chown', path, uid, gid)

    def truncate(self, path, length):
        return self.operations('truncate', path, length)

    def open(self, path, fip):
        fi = fip.contents
        if self.raw_fi:
            return self.operations('open', path, fi)
        else:
            fi.fh = self.operations('open', path, fi.flags)
            return 0

    def read(self, path, buf, size, offset, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        ret = self.operations('read', path, size, offset, fh)
        if not ret:
            return 0
        data = create_string_buffer(ret[:size], size)
        memmove(buf, data, size)
        return size

    def write(self, path, buf, size, offset, fip):
        data = string_at(buf, size)
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('write', path, data, offset, fh)

    def statfs(self, path, buf):
        stv = buf.contents
        attrs = self.operations('statfs', path)
        for key, val in attrs.items():
            if hasattr(stv, key):
                setattr(stv, key, val)
        return 0

    def flush(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('flush', path, fh)

    def release(self, path, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('release', path, fh)

    def fsync(self, path, datasync, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('fsync', path, datasync, fh)

    def setxattr(self, path, name, value, size, options, *args):
        data = string_at(value, size)
        return self.operations('setxattr', path, name, data, options, *args)

    def getxattr(self, path, name, value, size, *args):
        ret = self.operations('getxattr', path, name, *args)
        retsize = len(ret)
        buf = create_string_buffer(ret, retsize)    # Does not add trailing 0
        if bool(value):
            if retsize > size:
                return -ERANGE
            memmove(value, buf, retsize)
        return retsize

    def listxattr(self, path, namebuf, size):
        ret = self.operations('listxattr', path)
        buf = create_string_buffer('\x00'.join(ret)) if ret else ''
        bufsize = len(buf)
        if bool(namebuf):
            if bufsize > size:
                return -ERANGE
            memmove(namebuf, buf, bufsize)
        return bufsize

    def removexattr(self, path, name):
        return self.operations('removexattr', path, name)

    def opendir(self, path, fip):
        # Ignore raw_fi
        fip.contents.fh = self.operations('opendir', path)
        return 0

    def readdir(self, path, buf, filler, offset, fip):
        # Ignore raw_fi
        for item in self.operations('readdir', path, fip.contents.fh):
            if isinstance(item, str):
                name, st, offset = item, None, 0
                name = name.encode('utf-8')
            else:
                name, attrs, offset = item
                if attrs:
                    st = c_stat()
                    set_st_attrs(st, attrs)
                else:
                    st = None
            if filler(buf, name, st, offset) != 0:
                break
        return 0

    def releasedir(self, path, fip):
        # Ignore raw_fi
        return self.operations('releasedir', path, fip.contents.fh)

    def fsyncdir(self, path, datasync, fip):
        # Ignore raw_fi
        return self.operations('fsyncdir', path, datasync, fip.contents.fh)

    def init(self, conn):
        return self.operations('init', '/')

    def destroy(self, private_data):
        return self.operations('destroy', '/')

    def access(self, path, amode):
        return self.operations('access', path, amode)

    def create(self, path, mode, fip):
        fi = fip.contents
        if self.raw_fi:
            return self.operations('create', path, mode, fi)
        else:
            fi.fh = self.operations('create', path, mode)
            return 0

    def ftruncate(self, path, length, fip):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('truncate', path, length, fh)

    def fgetattr(self, path, buf, fip):
        memset(buf, 0, sizeof(c_stat))
        st = buf.contents
        fh = fip and (fip.contents if self.raw_fi else fip.contents.fh)
        attrs = self.operations('getattr', path, fh)
        set_st_attrs(st, attrs)
        return 0

    def lock(self, path, fip, cmd, lock):
        fh = fip.contents if self.raw_fi else fip.contents.fh
        return self.operations('lock', path, fh, cmd, lock)

    def utimens(self, path, buf):
        if buf:
            atime = time_of_timespec(buf.contents.actime)
            mtime = time_of_timespec(buf.contents.modtime)
            times = (atime, mtime)
        else:
            times = None
        return self.operations('utimens', path, times)

    def bmap(self, path, blocksize, idx):
        return self.operations('bmap', path, blocksize, idx)


class Operations(object):
    """This class should be subclassed and passed as an argument to FUSE on
       initialization. All operations should raise an OSError exception on
       error.

       When in doubt of what an operation should do, check the FUSE header
       file or the corresponding system call man page."""

    def __call__(self, op, *args):
        if not hasattr(self, op):
            raise OSError(EFAULT, '')
        return getattr(self, op)(*args)

    def access(self, path, amode):
        return 0

    bmap = None

    def chmod(self, path, mode):
        raise OSError(EROFS, '')

    def chown(self, path, uid, gid):
        raise OSError(EROFS, '')

    def create(self, path, mode, fi=None):
        """When raw_fi is False (default case), fi is None and create should
           return a numerical file handle.
           When raw_fi is True the file handle should be set directly by create
           and return 0."""
        raise OSError(EROFS, '')

    def destroy(self, path):
        """Called on filesystem destruction. Path is always /"""
        pass

    def flush(self, path, fh):
        return 0

    def fsync(self, path, datasync, fh):
        return 0

    def fsyncdir(self, path, datasync, fh):
        return 0

    def getattr(self, path, fh=None):
        """Returns a dictionary with keys identical to the stat C structure
           of stat(2).
           st_atime, st_mtime and st_ctime should be floats.
           NOTE: There is an incombatibility between Linux and Mac OS X concerning
           st_nlink of directories. Mac OS X counts all files inside the directory,
           while Linux counts only the subdirectories."""

        if path != '/':
            raise OSError(ENOENT, '')
        return dict(st_mode=(S_IFDIR | 0o755), st_nlink=2)

    def getxattr(self, path, name, position=0):
        raise OSError(ENOTSUP, '')

    def init(self, path):
        """Called on filesystem initialization. Path is always /
           Use it instead of __init__ if you start threads on initialization."""
        pass

    def link(self, target, source):
        raise OSError(EROFS, '')

    def listxattr(self, path):
        return []

    lock = None

    def mkdir(self, path, mode):
        raise OSError(EROFS, '')

    def mknod(self, path, mode, dev):
        raise OSError(EROFS, '')

    def open(self, path, flags):
        """When raw_fi is False (default case), open should return a numerical
           file handle.
           When raw_fi is True the signature of open becomes:
               open(self, path, fi)
           and the file handle should be set directly."""
        return 0

    def opendir(self, path):
        """Returns a numerical file handle."""
        return 0

    def read(self, path, size, offset, fh):
        """Returns a string containing the data requested."""
        raise OSError(ENOENT, '')

    def readdir(self, path, fh):
        """Can return either a list of names, or a list of (name, attrs, offset)
           tuples. attrs is a dict as in getattr."""
        return ['.', '..']

    def readlink(self, path):
        raise OSError(ENOENT, '')

    def release(self, path, fh):
        return 0

    def releasedir(self, path, fh):
        return 0

    def removexattr(self, path, name):
        raise OSError(ENOTSUP, '')

    def rename(self, old, new):
        raise OSError(EROFS, '')

    def rmdir(self, path):
        raise OSError(EROFS, '')

    def setxattr(self, path, name, value, options, position=0):
        raise OSError(ENOTSUP, '')

    def statfs(self, path):
        """Returns a dictionary with keys identical to the statvfs C structure
           of statvfs(3).
           On Mac OS X f_bsize and f_frsize must be a power of 2 (minimum 512)."""
        return {}

    def symlink(self, target, source):
        raise OSError(EROFS, '')

    def truncate(self, path, length, fh=None):
        raise OSError(EROFS, '')

    def unlink(self, path):
        raise OSError(EROFS, '')

    def utimens(self, path, times=None):
        """Times is a (atime, mtime) tuple. If None use current time."""
        return 0

    def write(self, path, data, offset, fh):
        raise OSError(EROFS, '')


class LoggingMixIn:
    def __call__(self, op, path, *args):
        logging.debug('-> %s %s %s', op, path, repr(args))
        ret = '[Unknown Error]'
        try:
            ret = getattr(self, op)(path, *args)
            return ret
        except OSError as e:
            ret = str(e)
            raise
        finally:
            logging.debug('<- %s %s', op, repr(ret))



#-------------------------------------------------------------------------------
#
#  Defines the IMutableTemplate interface used to define modifiable templates.
#
#  Written by: David C. Morrill
#
#  Date: 08/01/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the IMutableTemplate interface used to define modifiable templates.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Event

from itemplate \
    import ITemplate

#-------------------------------------------------------------------------------
#  'IMutableTemplate' interface:
#-------------------------------------------------------------------------------

class IMutableTemplate ( ITemplate ):
    """ Defines the IMutableTemplate interface used to define modifiable
        templates.

        Extends the ITemplate interface by supporting the ability for the
        template to modify itself.
    """

    # An event fired when the template mutates (i.e. changes in some way that
    # may affect the number of data sources it exposes, and so on):
    template_mutated = Event




#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Jan 2012
#
#------------------------------------------------------------------------------

"""
Test case for bug (wx, Mac OS X)

Editing the text part of a spin control box and pressing the OK button
without de-focusing raises an AttributeError

Traceback (most recent call last):
  File "ETS/traitsui/traitsui/wx/range_editor.py", line 783, in update_object
    self.value = self.control.GetValue()
AttributeError: 'NoneType' object has no attribute 'GetValue'
"""
from __future__ import print_function

from traits.has_traits import HasTraits
from traits.trait_types import Int
from traitsui.item import Item
from traitsui.view import View
from traitsui.editors.range_editor import RangeEditor

from traitsui.tests._tools import *


class NumberWithSpinnerEditor(HasTraits):
    """Dialog containing a RangeEditor in 'spinner' mode for an Int.
    """

    number = Int

    traits_view = View(
        Item(label="Enter 4, then press OK without defocusing"),
        Item('number', editor=RangeEditor(low=3, high=8, mode='spinner')),
        buttons=['OK']
    )


@skip_if_not_wx
def test_wx_spin_control_editing_should_not_crash():
    # Bug: when editing the text part of a spin control box, pressing
    # the OK button raises an AttributeError on Mac OS X

    try:
        with store_exceptions_on_all_threads():
            num = NumberWithSpinnerEditor()
            ui = num.edit_traits()

            # the following is equivalent to clicking in the text control of the
            # range editor, enter a number, and clicking ok without defocusing

            # SpinCtrl object
            spin = ui.control.FindWindowByName('wxSpinCtrl')
            spin.SetFocusFromKbd()

            # on Windows, a wxSpinCtrl does not have children, and we cannot do
            # the more fine-grained testing below
            if len(spin.GetChildren()) == 0:
                spin.SetValueString('4')
            else:
                # TextCtrl object of the spin control
                spintxt = spin.FindWindowByName('text')
                spintxt.SetValue('4')

            # press the OK button and close the dialog
            press_ok_button(ui)
    except AttributeError:
        # if all went well, we should not be here
        assert False, "AttributeError raised"


@skip_if_not_wx
def test_wx_spin_control_editing_does_not_update():
    # Bug: when editing the text part of a spin control box, pressing
    # the OK button does not update the value of the HasTraits class
    # on Mac OS X

    with store_exceptions_on_all_threads():
        num = NumberWithSpinnerEditor()
        ui = num.edit_traits()

        # the following is equivalent to clicking in the text control of the
        # range editor, enter a number, and clicking ok without defocusing

        # SpinCtrl object
        spin = ui.control.FindWindowByName('wxSpinCtrl')
        spin.SetFocusFromKbd()

        # on Windows, a wxSpinCtrl does not have children, and we cannot do
        # the more fine-grained testing below
        if len(spin.GetChildren()) == 0:
            spin.SetValueString('4')
        else:
            # TextCtrl object of the spin control
            spintxt = spin.FindWindowByName('text')
            spintxt.SetValue('4')

        # press the OK button and close the dialog
        press_ok_button(ui)

        # if all went well, the number traits has been updated and its value is
        # 4
        assert num.number == 4


@skip_if_not_qt4
def test_qt_spin_control_editing():
    # Behavior: when editing the text part of a spin control box, pressing
    # the OK button updates the value of the HasTraits class

    from pyface import qt

    with store_exceptions_on_all_threads():
        num = NumberWithSpinnerEditor()
        ui = num.edit_traits()

        # the following is equivalent to clicking in the text control of the
        # range editor, enter a number, and clicking ok without defocusing

        # text element inside the spin control
        lineedit = ui.control.findChild(qt.QtGui.QLineEdit)
        lineedit.setFocus()
        lineedit.setText('4')

        # press the OK button and close the dialog
        press_ok_button(ui)

    # if all went well, the number traits has been updated and its value is 4
    assert num.number == 4


if __name__ == '__main__':
    # Executing the file opens the dialog for manual testing
    num = NumberWithSpinnerEditor()
    num.configure_traits()
    print(num.number)



#!/usr/bin/env python

from fs.errors import FSError
from fs.opener import opener
from fs.path import pathsplit, abspath, isdotfile, iswildcard
from fs.commands.runner import Command
from collections import defaultdict
import sys

class FSls(Command):

    usage = """fsls [OPTIONS]... [PATH]
List contents of [PATH]"""


    def get_optparse(self):
        optparse = super(FSls, self).get_optparse()
        optparse.add_option('-u', '--full', dest='fullpath', action="store_true", default=False,
                            help="output full path", metavar="FULL")
        optparse.add_option('-s', '--syspath', dest='syspath', action="store_true", default=False,
                            help="output system path (if one exists)", metavar="SYSPATH")
        optparse.add_option('-r', '--url', dest='url', action="store_true", default=False,
                            help="output URL in place of path (if one exists)", metavar="URL")
        optparse.add_option('-d', '--dirsonly', dest='dirsonly', action="store_true", default=False,
                            help="list directories only", metavar="DIRSONLY")
        optparse.add_option('-f', '--filesonly', dest='filesonly', action="store_true", default=False,
                            help="list files only", metavar="FILESONLY")
        optparse.add_option('-l', '--long', dest='long', action="store_true", default=False,
                            help="use a long listing format", metavar="LONG")
        optparse.add_option('-a', '--all', dest='all', action='store_true', default=False,
                            help="do not hide dot files")

        return optparse


    def do_run(self, options, args):
        output = self.output

        if not args:
            args = [u'.']

        dir_paths = []
        file_paths = []
        fs_used = set()
        for fs_url in args:
            fs, path = self.open_fs(fs_url)
            fs_used.add(fs)
            path = path or '.'
            wildcard = None

            if iswildcard(path):
                path, wildcard = pathsplit(path)

            if path != '.' and fs.isfile(path):
                if not options.dirsonly:
                    file_paths.append(path)
            else:
                if not options.filesonly:
                    dir_paths += fs.listdir(path,
                                            wildcard=wildcard,
                                            full=options.fullpath or options.url,
                                            dirs_only=True)

                if not options.dirsonly:
                    file_paths += fs.listdir(path,
                                             wildcard=wildcard,
                                             full=options.fullpath or options.url,
                                             files_only=True)

        for fs in fs_used:
            try:
                fs.close()
            except FSError:
                pass

        if options.syspath:
            # Path without a syspath, just won't be displayed
            dir_paths = filter(None, [fs.getsyspath(path, allow_none=True) for path in dir_paths])
            file_paths = filter(None, [fs.getsyspath(path, allow_none=True) for path in file_paths])

        if options.url:
            # Path without a syspath, just won't be displayed
            dir_paths = filter(None, [fs.getpathurl(path, allow_none=True) for path in dir_paths])
            file_paths = filter(None, [fs.getpathurl(path, allow_none=True) for path in file_paths])

        dirs = frozenset(dir_paths)
        paths = sorted(dir_paths + file_paths, key=lambda p: p.lower())

        if not options.all:
            paths = [path for path in paths if not isdotfile(path)]

        if not paths:
            return

        def columnize(paths, num_columns):

            col_height = (len(paths) + num_columns - 1) / num_columns
            columns = [[] for _ in xrange(num_columns)]
            col_no = 0
            col_pos = 0
            for path in paths:
                columns[col_no].append(path)
                col_pos += 1
                if col_pos >= col_height:
                    col_no += 1
                    col_pos = 0

            padded_columns = []

            wrap_filename = self.wrap_filename
            wrap_dirname = self.wrap_dirname

            def wrap(path):
                if path in dirs:
                    return wrap_dirname(path.ljust(max_width))
                else:
                    return wrap_filename(path.ljust(max_width))

            for column in columns:
                if column:
                    max_width = max([len(path) for path in column])
                else:
                    max_width = 1
                max_width = min(max_width, terminal_width)
                padded_columns.append([wrap(path) for path in column])

            return padded_columns

        def condense_columns(columns):
            max_column_height = max([len(col) for col in columns])
            lines = [[] for _ in xrange(max_column_height)]
            for column in columns:
                for line, path in zip(lines, column):
                    line.append(path)
            return '\n'.join(u'  '.join(line) for line in lines)

        if options.long:
            for path in paths:
                if path in dirs:
                    output((self.wrap_dirname(path), '\n'))
                else:
                    output((self.wrap_filename(path), '\n'))

        else:
            terminal_width = self.terminal_width
            path_widths = [len(path) for path in paths]
            smallest_paths = min(path_widths)
            num_paths = len(paths)

            num_cols = min(terminal_width // (smallest_paths + 2), num_paths)
            while num_cols:
                col_height = (num_paths + num_cols - 1) // num_cols
                line_width = 0
                for col_no in xrange(num_cols):
                    try:
                        col_width = max(path_widths[col_no * col_height: (col_no + 1) * col_height])
                    except ValueError:
                        continue
                    line_width += col_width
                    if line_width > terminal_width:
                        break
                    line_width += 2
                else:
                    if line_width - 1 <= terminal_width:
                        break
                num_cols -= 1
            num_cols = max(1, num_cols)
            columns = columnize(paths, num_cols)
            output((condense_columns(columns), '\n'))

def run():
    return FSls().run()

if __name__ == "__main__":
    sys.exit(run())



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Jan 2012
#
#------------------------------------------------------------------------------

from functools import partial
from contextlib import contextmanager
import nose

import sys
import traceback

from traits.etsconfig.api import ETSConfig
import traits.trait_notifiers

# ######### Testing tools


@contextmanager
def store_exceptions_on_all_threads():
    """Context manager that captures all exceptions, even those coming from
    the UI thread. On exit, the first exception is raised (if any).

    It also temporarily overwrites the global function
    traits.trait_notifier.handle_exception , which logs exceptions to
    console without re-raising them by default.
    """

    exceptions = []

    def _print_uncaught_exception(type, value, tb):
        message = 'Uncaught exception:\n'
        message += ''.join(traceback.format_exception(type, value, tb))
        print message

    def excepthook(type, value, tb):
        exceptions.append(value)
        _print_uncaught_exception(type, value, tb)

    def handle_exception(object, trait_name, old, new):
        type, value, tb = sys.exc_info()
        exceptions.append(value)
        _print_uncaught_exception(type, value, tb)

    _original_handle_exception = traits.trait_notifiers.handle_exception
    try:
        sys.excepthook = excepthook
        traits.trait_notifiers.handle_exception = handle_exception
        yield
    finally:
        if len(exceptions) > 0:
            raise exceptions[0]
        sys.excepthook = sys.__excepthook__
        traits.trait_notifiers.handle_exception = _original_handle_exception


def _is_current_backend(backend_name=''):
    return ETSConfig.toolkit == backend_name


def skip_if_not_backend(test_func, backend_name=''):
    """Decorator that skip tests if the backend is not the desired one."""

    if not _is_current_backend(backend_name):
        # preserve original name so that it appears in the report
        orig_name = test_func.__name__

        def test_func():
            raise nose.SkipTest
        test_func.__name__ = orig_name

    return test_func


#: Return True if current backend is 'wx'
is_current_backend_wx = partial(_is_current_backend, backend_name='wx')

#: Return True if current backend is 'qt4'
is_current_backend_qt4 = partial(_is_current_backend, backend_name='qt4')

#: Return True if current backend is 'null'
is_current_backend_null = partial(_is_current_backend, backend_name='null')


#: Test decorator: Skip test if backend is not 'wx'
skip_if_not_wx = partial(skip_if_not_backend, backend_name='wx')

#: Test decorator: Skip test if backend is not 'qt4'
skip_if_not_qt4 = partial(skip_if_not_backend, backend_name='qt4')

#: Test decorator: Skip test if backend is not 'null'
skip_if_not_null = partial(skip_if_not_backend, backend_name='null')


def skip_if_null(test_func):
    """Decorator that skip tests if the backend is set to 'null'.

    Some tests handle both wx and Qt in one go, but many things are not
    defined in the null backend. Use this decorator to skip the test.
    """

    if _is_current_backend('null'):
        # preserve original name so that it appears in the report
        orig_name = test_func.__name__

        def test_func():
            raise nose.SkipTest
        test_func.__name__ = orig_name

    return test_func


def count_calls(func):
    """Decorator that stores the number of times a function is called.

    The counter is stored in func._n_counts.
    """

    def wrapped(*args, **kwargs):
        wrapped._n_calls += 1
        return func(*args, **kwargs)

    wrapped._n_calls = 0

    return wrapped


# ######### Utility tools to test on both qt4 and wx

def get_children(node):
    if ETSConfig.toolkit == 'wx':
        return node.GetChildren()
    else:
        return node.children()


def press_ok_button(ui):
    """Press the OK button in a wx or qt dialog."""

    if is_current_backend_wx():
        import wx

        ok_button = ui.control.FindWindowByName('button')
        click_event = wx.CommandEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED,
                                      ok_button.GetId())
        ok_button.ProcessEvent(click_event)

    elif is_current_backend_qt4():
        from pyface import qt

        # press the OK button and close the dialog
        ok_button = ui.control.findChild(qt.QtGui.QPushButton)
        ok_button.click()


def get_dialog_size(ui_control):
    """Return the size of the dialog.

    Return a tuple (width, height) with the size of the dialog in pixels.
    E.g.:

        >>> get_dialog_size(ui.control)
    """

    if is_current_backend_wx():
        return ui_control.GetSizeTuple()

    elif is_current_backend_qt4():
        return ui_control.size().width(), ui_control.size().height()


# ######### Debug tools

def apply_on_children(func, node, _level=0):
    """Print the result of applying a function on `node` and its children.
    """
    print '-' * _level + str(node)
    print ' ' * _level + str(func(node)) + '\n'
    for child in get_children(node):
        apply_on_children(func, child, _level + 1)


def wx_print_names(node):
    """Print the name and id of `node` and its children.

    Use as::

        >>> ui = xxx.edit_traits()
        >>> wx_print_names(ui.control)
    """
    apply_on_children(lambda n: (n.GetName(), n.GetId()), node)


def qt_print_names(node):
    """Print the name of `node` and its children.

    Use as::

        >>> ui = xxx.edit_traits()
        >>> qt_print_names(ui.control)
    """
    apply_on_children(lambda n: n.objectName(), node)


def wx_announce_when_destroyed(node):
    """Prints a message when `node` is destroyed.

    Use as:

        >>> ui = xxx.edit_traits()
        >>> apply_on_children(wx_announce_when_destroyed, ui.control)
    """

    _destroy_method = node.Destroy

    def destroy_wrapped():
        print 'Destroying:', node
        #print 'Stack is'
        #traceback.print_stack()
        _destroy_method()
        print 'Destroyed:', node

    node.Destroy = destroy_wrapped
    return 'Node {} decorated'.format(node.GetName())


def wx_find_event_by_number(evt_num):
    """Find all wx event names that correspond to a certain event number.

    Example:

        >>> wx_find_event_by_number(10010)
        ['wxEVT_COMMAND_MENU_SELECTED', 'wxEVT_COMMAND_TOOL_CLICKED']
    """

    import wx
    possible = [attr for attr in dir(wx)
                if attr.startswith('wxEVT') and getattr(wx, attr) == evt_num]

    return possible



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/06/2005
#
#------------------------------------------------------------------------------

""" Pseudo-package for all of the core symbols from Traits and TraitsUI.
Use this module for importing Traits names into your namespace. For example::

    from traits.api import HasTraits
"""

from __future__ import absolute_import

from .trait_base import Uninitialized, Undefined, Missing, Self, python_version

from .trait_errors import TraitError, TraitNotificationError, DelegationError

from .trait_notifiers import (push_exception_handler, pop_exception_handler,
        TraitChangeNotifyWrapper)

from .category import Category

from .traits import (CTrait, Trait, Property, TraitFactory, Default, Color,
        RGBColor, Font)

from .trait_types import (Any, Generic, Int, Long, Float, Complex, Str, Title,
        Unicode, Bool, CInt, CLong, CFloat, CComplex, CStr, CUnicode, CBool,
        String, Regex, Code, HTML, Password, Callable, This, self, Function,
        Method, Module, Python, ReadOnly, Disallow, Constant,
        Delegate, DelegatesTo, PrototypedFrom, Expression, PythonValue, File,
        Directory, Range, Enum, Tuple, List, CList, Set, CSet, Dict, Instance,
        AdaptedTo, AdaptsTo, Event, Button, ToolbarButton, Either, Type,
        Symbol, WeakRef, Date, Time, false, true, undefined, Supports)

from .trait_types import (ListInt, ListFloat, ListStr, ListUnicode,
        ListComplex, ListBool, ListFunction, ListMethod,
        ListThis, DictStrAny, DictStrStr, DictStrInt,
        DictStrLong, DictStrFloat, DictStrBool, DictStrList)

try:
    from .trait_types import Class, ListClass, ListInstance
except ImportError:
    # Python 3 does not have old-style classes anymore, so Class does not exist
    # interestingly, ListInstance is not equivalent to List(Instance), but
    # rather only allows old-style instances.
    pass

from .trait_types import (BaseInt, BaseLong, BaseFloat, BaseComplex, BaseStr,
        BaseUnicode, BaseBool, BaseCInt, BaseCLong, BaseCFloat, BaseCComplex,
        BaseCStr, BaseCUnicode, BaseCBool, BaseFile, BaseDirectory, BaseRange,
        BaseEnum, BaseTuple, BaseInstance)

from .trait_types import UUID, ValidatedTuple

from .has_traits import (HasTraits, HasStrictTraits, HasPrivateTraits,
        Interface, SingletonHasTraits, SingletonHasStrictTraits,
        SingletonHasPrivateTraits, MetaHasTraits, Vetoable, VetoableEvent,
        implements, traits_super, on_trait_change, cached_property,
        property_depends_on, provides, isinterface)

try:
    from .has_traits import ABCHasTraits, ABCHasStrictTraits, ABCMetaHasTraits
except ImportError:
    pass

from .trait_handlers import (BaseTraitHandler, TraitType, TraitHandler,
        TraitRange, TraitString, TraitCoerceType, TraitCastType, TraitInstance,
        ThisClass, TraitClass, TraitFunction, TraitEnum, TraitPrefixList,
        TraitMap, TraitPrefixMap, TraitCompound, TraitList, TraitListObject,
        TraitListEvent, TraitSetObject, TraitSetEvent, TraitDict,
        TraitDictObject, TraitDictEvent, TraitTuple, NO_COMPARE,
        OBJECT_IDENTITY_COMPARE, RICH_COMPARE)

from .trait_value import (BaseTraitValue, TraitValue, SyncValue,
        TypeValue, DefaultValue)

from .adaptation.adapter import Adapter, adapts
from .adaptation.adaptation_error import AdaptationError
from .adaptation.adaptation_manager import adapt, register_factory, \
     register_provides

from .trait_numeric import Array, ArrayOrNone, CArray

try:
    from . import has_traits as has_traits
    #---------------------------------------------------------------------------
    #  Patch the main traits module with the correct definition for the
    #  ViewElements class:
    #  NOTE: We do this in a try..except block because traits.ui depends on
    #  the pyface module (part of the TraitsGUI package) which may not
    #  necessarily be installed. Not having TraitsGUI means that the 'ui'
    #  features of traits will not work.
    #---------------------------------------------------------------------------

    from traitsui import view_elements
    has_traits.ViewElements = view_elements.ViewElements

    #-------------------------------------------------------------------------------
    #  Patch the main traits module with the correct definition for the
    #  ViewElement and ViewSubElement class:
    #-------------------------------------------------------------------------------

    has_traits.ViewElement = view_elements.ViewElement
except ImportError:
    pass



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A menu that mimics the standard MDI window menu.

This is the menu that has the tile/cascade actions etc.

"""


# Enthought library imports.
from traits.api import Str

# Local imports.
from action.api import MenuManager, Separator, WindowAction


class Cascade(WindowAction):
    """ Cascades the windows. """

    #### 'Action' interface ###################################################

    name    = Str("Ca&scade")
    tooltip = Str("Cascade the windows")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Cascades the windows. """

        self.window.control.Cascade()

        return

class Tile(WindowAction):
    """ Tiles the windows horizontally. """

    #### 'Action' interface ###################################################

    name    = Str("&Tile")
    tooltip = Str("Tile the windows")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Tiles the windows horizontally. """

        self.window.control.Tile()

        return

class ArrangeIcons(WindowAction):
    """ Arranges the icons. """

    #### 'Action' interface ###################################################

    name    = Str("&Arrange Icons")
    tooltip = Str("Arrange the icons")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Arranges the icons. """

        self.window.control.ArrangeIcons()

        return

class Next(WindowAction):
    """ Activates the next window. """

    #### 'Action' interface ###################################################

    name    = Str("&Next")
    tooltip = Str("Activate the next window")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Activates the next window. """

        self.window.control.ActivateNext()

        return

class Previous(WindowAction):
    """ Activates the previous window. """

    #### 'Action' interface ###################################################

    name    = Str("&Previous")
    tooltip = Str("Activate the previous window")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Activates the previous window. """

        self.window.control.ActivatePrevious()

        return

class Close(WindowAction):
    """ Closes the current window. """

    #### 'Action' interface ###################################################

    name    = Str("&Close")
    tooltip = Str("Close the current window")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Closes the current window. """

        page = self.window.control.GetActiveChild()
        if page is not None:
            page.Close()

        return

class CloseAll(WindowAction):
    """ Closes all of the child windows. """

    #### 'Action' interface ###################################################

    name    = Str("Close A&ll")
    tooltip = Str("Close all of the windows.")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Closes the child windows. """

        for page in self.window.control.GetChildren():
            page.Close()

        return

class MDIWindowMenu(MenuManager):
    """ A menu that mimics the standard MDI window menus.

    This is the menu that has the tile/cascade actions etc.

    """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, window):
        """ Creates a new MDI window menu. """

        # Base class constructor.
        super(MDIWindowMenu, self).__init__(
            Cascade(window=window),
            Tile(window=window),
            Separator(),
            ArrangeIcons(window=window),
            Next(window=window),
            Previous(window=window),
            Close(window=window),
            CloseAll(window=window),
            name = '&Window'
        )

#### EOF ######################################################################



"""Image editor sample task

"""
# Enthought library imports.
from pyface.api import ImageResource
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import DockPaneToggleGroup, SMenuBar, \
    SMenu, SToolBar, TaskAction, TaskToggleGroup
from traits.api import on_trait_change, Property, Instance


from omnivore.framework.task import FrameworkTask
from panes import Pane1, Pane2, Pane3
from image_editor import ImageEditor
from preferences import ImageEditPreferences


class ImageEditTask(FrameworkTask):
    """ A simple task for opening a blank editor.
    """

    new_file_text = ""

    #### Task interface #######################################################

    id = 'omnivore.framework.image_edit_task'
    name = 'Image Editor'

    preferences_helper = ImageEditPreferences

    ###########################################################################
    # 'Task' interface.
    ###########################################################################

    def _default_layout_default(self):
        return TaskLayout(
            top=HSplitter(
                PaneItem('image_edit.pane1'),
                PaneItem('image_edit.pane2'),
                PaneItem('image_edit.pane3'),
                ))

    def create_dock_panes(self):
        """ Create the file browser and connect to its double click event.
        """
        return [ Pane1(), Pane2(), Pane3() ]

    ###########################################################################
    # 'FrameworkTask' interface.
    ###########################################################################

    def get_editor(self, guess=None):
        """ Opens a new empty window
        """
        editor = ImageEditor()
        return editor

    ###
    @classmethod
    def can_edit(cls, document):
        mime = document.metadata.mime
        return mime == "image/jpeg" or mime == "image/png"



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------

from __future__ import absolute_import

# Enthought library imports.
from traits.api import Any, HasTraits, Unicode, provides

# Local imports.
from .i_command import ICommand


@provides(ICommand)
class AbstractCommand(HasTraits):
    """ The AbstractCommand class is an abstract base class that implements the
    ICommand interface.
    """

    #### 'ICommand' interface #################################################

    # This is the data on which the command operates.
    data = Any

    # This is the name of the command as it will appear in any GUI element.  It
    # may include '&' which will be automatically removed whenever it is
    # inappropriate.
    name = Unicode

    ###########################################################################
    # 'ICommand' interface.
    ###########################################################################

    def do(self):
        """ This is called by the command stack to do the command and to return
        any value.  The command must save any state necessary for the 'redo()'
        and 'undo()' methods to work.  The class's __init__() must also ensure
        that deep copies of any arguments are made if appropriate.  It is
        guaranteed that this will only ever be called once and that it will be
        called before any call to 'redo()' or 'undo()'.
        """

        raise NotImplementedError

    def merge(self, other):
        """ This is called by the command stack to try and merge another
        command with this one.  True is returned if the commands were merged.
        'other' is the command that is about to be executed.  If the commands
        are merged then 'other' will discarded and not placed on the command
        stack.  A subsequent undo or redo of this modified command must have
        the same effect as the two original commands.
        """

        # By default merges never happen.
        return False

    def redo(self):
        """ This is called by the command stack to redo the command.  Any
        returned value will replace the value that the command stack references
        from the original call to 'do()' or previous call to 'redo()'.
        """

        raise NotImplementedError

    def undo(self):
        """ This is called by the command stack to undo the command. """

        raise NotImplementedError



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   08/18/2008
#
#------------------------------------------------------------------------------

""" Defines some helper classes and traits used to define 'bindable' editor
    values.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import HasPrivateTraits, Instance, Str, Int, Float, Either

#-------------------------------------------------------------------------
#  'ContextValue' class:
#-------------------------------------------------------------------------


class ContextValue(HasPrivateTraits):
    """ Defines the name of a context value that can be bound to some editor
        value.
    """

    # The extended trait name of the value that can be bound to the editor
    # (e.g. 'selection' or 'handler.selection'):
    name = Str

    #-- object Interface -----------------------------------------------------

    def __init__(self, name):
        """ Initializes the object.
        """
        self.name = name

# Define a shorthand name for a ContextValue:
CV = ContextValue

#-------------------------------------------------------------------------
#  Trait definitions useful in defining bindable editor traits:
#-------------------------------------------------------------------------

InstanceOfContextValue = Instance(ContextValue, allow_none=False)


def CVType(type):
    return Either(type, InstanceOfContextValue, sync_value='to')

CVInt = CVType(Int)
CVFloat = CVType(Float)
CVStr = CVType(Str)



"""
fs.httpfs
=========


"""

from fs.base import FS
from fs.path import normpath
from fs.errors import ResourceNotFoundError, UnsupportedError
from fs.filelike import FileWrapper
from fs import iotools

from urllib2 import urlopen, URLError
from datetime import datetime


class HTTPFS(FS):

    """Can barely be called a filesystem, because HTTP servers generally don't support
    typical filesystem functionality. This class exists to allow the :doc:`opener` system
    to read files over HTTP.

    If you do need filesystem like functionality over HTTP, see :mod:`fs.contrib.davfs`.

    """

    _meta = {'read_only': True,
             'network': True}

    def __init__(self, url):
        """

        :param url: The base URL

        """
        self.root_url = url

    def _make_url(self, path):
        path = normpath(path)
        url = '%s/%s' % (self.root_url.rstrip('/'), path.lstrip('/'))
        return url

    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):

        if '+' in mode or 'w' in mode or 'a' in mode:
            raise UnsupportedError('write')

        url = self._make_url(path)
        try:
            f = urlopen(url)
        except URLError, e:
            raise ResourceNotFoundError(path, details=e)
        except OSError, e:
            raise ResourceNotFoundError(path, details=e)

        return FileWrapper(f)

    def exists(self, path):
        return self.isfile(path)

    def isdir(self, path):
        return False

    def isfile(self, path):
        url = self._make_url(path)
        f = None
        try:
            try:
                f = urlopen(url)
            except (URLError, OSError):
                return False
        finally:
            if f is not None:
                f.close()

        return True

    def listdir(self, path="./",
                      wildcard=None,
                      full=False,
                      absolute=False,
                      dirs_only=False,
                      files_only=False):
        return []

    def getinfo(self, path):
        url = self._make_url(path)
        info = urlopen(url).info().dict
        if 'content-length' in info:
            info['size'] = info['content-length']
        if 'last-modified' in info:
            info['modified_time'] = datetime.strptime(info['last-modified'],
                                                      "%a, %d %b %Y %H:%M:%S %Z")
        return info



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005, 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#  Author: Duncan Child <duncan@enthought.com>
#
#-----------------------------------------------------------------------------

""" A record of refactorings to be performed during unpickling of objects.
"""

# Standard library imports
import logging

# Enthought library imports
from traits.api import Dict, HasPrivateTraits, Int, List, Tuple, Str


logger = logging.getLogger(__name__)


##############################################################################
# class 'Updater'
##############################################################################

class Updater(HasPrivateTraits):
    """ A record of refactorings to be performed during unpickling of objects.
    """

    ##########################################################################
    # Traits
    ##########################################################################

    ### public 'Updater' interface ###########################################

    # Mappings from a pickled class to a class it should be unpickled as.
    #
    # The keys are a tuple of the source class's module and class names in
    # that order.  The values are the target class's module and class names
    # in that order.
    class_map = Dict(Tuple(Str, Str), Tuple(Str, Str))

    # State functions that should be called to convert state from one version
    # of a class to another.
    #
    # The keys are a tuple of the class's module name, class name, and version
    # in that order.  The values are a list of functions to be called during
    # unpickling to do the state conversion.   Note that the version in the
    # key represents the version that the function converts *TO*.
    state_functions = Dict(Tuple(Str, Str, Int), List)

    # Our record of the attribute that records the version number for a
    # specific class.  If no record is found for a given class, then the
    # default value is used instead -- see '_default_version_attribute'.
    #
    # The key is a tuple of the class's module name and class name in that
    # order.  The value is the name of the version attribute.
    version_attribute_map = Dict(Tuple(Str, Str), Str)


    ### protected 'Updater' interface ########################################

    # The default name of the attribute that declares the version of a class
    # or class instance.
    _default_version_attribute = '_enthought_pickle_version'

    # A record of which classes we have state functions for.
    #
    # The keys are a tuple of the class's module name and class name in that
    # order.  The values are reference counts.
    _state_function_classes = Dict(Tuple(Str, Str), Int)


    ##########################################################################
    # 'Updater' interface
    ##########################################################################

    ### public interface #####################################################

    def add_mapping(self, source_module, source_name, target_module,
        target_name):
        """ Adds a mapping from the class with the source name in the source
            module to the class with the target name in the target module.
        """
        self.class_map[(source_module, source_name)] = (target_module,
            target_name)


    def add_mapping_to_class(self, source_module, source_name, target_class):
        """ Convenience method to add a mapping, from the class with the
            source name in the source module to the target class.
        """
        self.add_mapping(source_module, source_name, target_class.__module__,
            target_class.__name__)


    def add_mappings(self, source_module, target_module, class_names):
        """ Adds mappings, from the specified source module to the specified
            target module, for each of the class names in the specified
            list.
        """
        for name in class_names:
            self.add_mapping(source_module, name, target_module, name)


    def add_state_function(self, module, name, target_version, function):
        """ Adds the specified function as a state function to be called to
            convert an instance of the class with the specified name within
            the specified module *TO* the specified version.

            Note that the framework handles calling of state functions to make
            the smallest version jumps possible.
        """
        key = (module, name, target_version)
        list = self.state_functions.setdefault(key, [])
        list = list[:] # Copy necessary because traits only recognizes list
                       # changes by list instance - not its contents.
        list.append(function)
        self.state_functions[key] = list


    def add_state_function_for_class(self, klass, target_version, function):
        """ Convenience method to add the specified function as a state
            function to be called to convert an instance of the specified
            class *TO* the specified version.
        """
        self.add_state_function(klass.__module__, klass.__name__,
            target_version, function)


    def declare_version_attribute(self, module, name, attribute_name):
        """ Adds the specified attribute name as the version attribute for the
            class within the specified module with the specified name.
        """
        self.version_attribute_map[(module, name)] = attribute_name


    def declare_version_attribute_for_class(self, klass, attribute_name):
        """ Covenience method to add the specified attribute name as the
            version attribute for the specified class.
        """
        self.declare_version_attribute(klass.__module__, klass.__name__,
            attribute_name)


    def get_version_attribute(self, module, name):
        """ Returns the name of the version attribute for the class of the
            specified name within the specified module.
        """
        return self.version_attribute_map.get( (module, name),
            self._default_version_attribute)


    def has_class_mapping(self, module, name):
        """ Returns True if this updater contains a class mapping for
            the class identified by the specified module and class name.
        """
        return (module, name) in self.class_map


    def has_state_function(self, module, name):
        """ Returns True if this updater contains any state functions for
            the class identified by the specified module and class name.
        """
        return (module, name) in self._state_function_classes


    def merge_updater(self, updater):
        """ Merges the mappings and state functions from the specified updater
            into this updater.
        """
        self.class_map.update(updater.class_map)
        self.version_attribute_map.update(updater.version_attribute_map)

        # The state functions dictionary requires special processing because
        # each value is a list and we don't just want to replace the existing
        # list with only the new content.
        for key, value in updater.state_functions.items():
            if isinstance(value, list) and len(value) > 0:
                funcs = self.state_functions.setdefault(key, [])
                funcs = funcs[:] # Copy necessary because traits only recognizes
                               # funcs changes by funcs instance - not its
                               # contents.
                funcs.extend(value)
                self.state_functions[key] = funcs


    ### trait handlers #######################################################

    def _class_map_changed(self, old, new):
        logger.debug('Detected class_map change from [%s] to [%s] in [%s]',
            old, new, self)


    def _class_map_items_changed(self, event):
        for o in event.removed:
            logger.debug('Detected [%s] removed from class_map in [%s]', o,
                self)
        for k, v in event.changed.items():
            logger.debug('Detected [%s] changed from [%s] to [%s] in ' + \
                'class_map in [%s]', k, v, self.class_map[k], self)
        for k, v in event.added.items():
            logger.debug('Detected mapping from [%s] to [%s] added to ' + \
                'class_map in [%s]', k, v, self)


    def _state_functions_changed(self, old, new):
        logger.debug('Detected state_functions changed from [%s] to [%s] ' + \
            'in [%s]', old, new, self)

        # Update our record of which classes we have state functions for.
        # All of our old state functions are gone so we simply need to rescan
        # the new functions.
        self._state_function_classes.clear()
        for key, value in new.items():
            module, name, version = key
            klass_key = (module, name)
            count = self._state_function_classes.setdefault(klass_key, 0)
            self._state_function_classes[klass_key] = count + len(value)


    def _state_functions_items_changed(self, event):
        # Decrement our reference counts for the classes we no longer
        # have state functions for.  If the reference count reaches zero,
        # remove the record completely.
        for k, v in event.removed.items():
            logger.debug('Detected [%s] removed from state_functions in [%s]',
                k, self)

            # Determine the new reference count of state functions for the
            # class who the removed item was for.
            module, name, version = k
            key = (module, name)
            count = self._state_function_classes[key] - len(v)

            # Store the new reference count.  Delete the entry if it is zero.
            if count < 0:
                logger.warn('Unexpectedly reached negative reference count ' +
                    'value of [%s] for [%s]', count, key)
                del self._state_function_classes[key]
            elif count == 0:
                del self._state_function_classes[key]
            else:
                self._state_function_classes[key] = count

        # Update our reference counts for changes to the list of functions
        # for a specific class and version.  The 'changed' dictionary's values
        # are the old values.
        for k, v in event.changed.items():
            value = self.state_functions[k]
            logger.debug('Detected [%s] changed in state_functions from ' + \
                '[%s] to [%s] in [%s]', k, v, value, self)

            # Determine the new reference count as a result of the change.
            module, name, version = k
            key = (module, name)
            count = self._state_function_classes[key] - len(v) + len(value)

            # Store the new reference count.  Delete the entry if it is zero.
            if count < 0:
                logger.warn('Unexpectedly reached negative reference count ' +
                    'value of [%s] for [%s]', count, key)
                del self._state_function_classes[key]
            elif count == 0:
                del self._state_function_classes[key]
            else:
                self._state_function_classes[key] = count


        # Update our reference counts for newly registered state functions.
        for k, v in event.added.items():
            logger.debug('Detected mapping of [%s] to [%s] added to ' + \
                'state_functions in [%s]', k, v, self)

            # Determine the new reference count as a result of the change.
            module, name, version = k
            key = (module, name)
            count = self._state_function_classes.setdefault(key, 0) + len(v)

            # Store the new reference count
            self._state_function_classes[key] = count


    def _version_attribute_map_changed(self, old, new):
        logger.debug('Detected version_attribute_map change from [%s] ' + \
            'to [%s] in [%s]', old, new, self)


    def _version_attribute_map_items_changed(self, event):
        for o in event.removed:
            logger.debug('Detected [%s] removed from version_attribute_map ' + \
                'in [%s]', o, self)
        for o in event.changed:
            logger.debug('Detected [%s] changed in version_attribute_map ' + \
                'in [%s]', o, self)
        for o in event.added:
            logger.debug('Detected [%s] added to version_attribute_map in ' + \
                '[%s]', o, self)


### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/22/2004
#
#------------------------------------------------------------------------------

""" Defines a subclass of the base wxPython color editor factory, for colors
    that are represented as tuples of the form ( *red*, *green*, *blue* ),
    where *red*, *green* and *blue* are floats in the range from 0.0 to 1.0.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.trait_base \
    import SequenceTypes

# Note: The ToolkitEditorFactory class imported from color_editor is the
# abstract ToolkitEditorFactory class (in traitsui.api) along with
# wx-specific methods added via a category. We need to override the
# implementations of the wx-specific methods here.
from color_editor \
    import ToolkitEditorFactory as BaseColorToolkitEditorFactory

#---------------------------------------------------------------------------
#  The wxPython ToolkitEditorFactory class.
#---------------------------------------------------------------------------


class ToolkitEditorFactory(BaseColorToolkitEditorFactory):
    """ wxPython editor factory for color editors.
    """
    #-------------------------------------------------------------------------
    #  Gets the wxPython color equivalent of the object:
    #-------------------------------------------------------------------------

    def to_wx_color(self, editor, color=None):
        """ Gets the wxPython color equivalent of the object trait.
        """
        if color is None:
            try:
                color = getattr(editor.object, editor.name + '_')
            except AttributeError:
                color = getattr(editor.object, editor.name)
        if isinstance(color, tuple):
            return wx.Colour(int(color[0] * 255.0),
                             int(color[1] * 255.0),
                             int(color[2] * 255.0))
        return color

    #-------------------------------------------------------------------------
    #  Gets the application equivalent of a wxPython value:
    #-------------------------------------------------------------------------

    def from_wx_color(self, color):
        """ Gets the application equivalent of a wxPython value.
        """
        return (color.Red() / 255.0,
                color.Green() / 255.0,
                color.Blue() / 255.0)

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified color value:
    #-------------------------------------------------------------------------

    def str_color(self, color):
        """ Returns the text representation of a specified color value.
        """
        if type(color) in SequenceTypes:
            return "(%d,%d,%d)" % (int(color[0] * 255.0),
                                   int(color[1] * 255.0),
                                   int(color[2] * 255.0))
        return super(ToolkitEditorFactory, self).str_color(color)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Any, Event, Property, provides, Unicode
from traits.api import Tuple

# Local imports.
from pyface.i_window import IWindow, MWindow
from pyface.key_pressed_event import KeyPressedEvent
from system_metrics import SystemMetrics
from .widget import Widget


@provides(IWindow)
class Window(MWindow, Widget):
    """ The toolkit specific implementation of a Window.  See the IWindow
    interface for the API documentation.
    """


    #### 'IWindow' interface ##################################################

    position = Property(Tuple)

    size = Property(Tuple)

    title = Unicode

    #### Events #####

    activated = Event

    closed =  Event

    closing =  Event

    deactivated = Event

    key_pressed = Event(KeyPressedEvent)

    opened = Event

    opening = Event

    #### Private interface ####################################################

    # Shadow trait for position.
    _position = Tuple((-1, -1))

    # Shadow trait for size.
    _size = Tuple((-1, -1))

    ###########################################################################
    # 'IWindow' interface.
    ###########################################################################

    def activate(self):
        self.control.Iconize(False)
        self.control.Raise()

    def show(self, visible):
        self.control.Show(visible)

    ###########################################################################
    # Protected 'IWindow' interface.
    ###########################################################################

    def _add_event_listeners(self):
        wx.EVT_ACTIVATE(self.control, self._wx_on_activate)
        wx.EVT_CLOSE(self.control, self._wx_on_close)
        wx.EVT_SIZE(self.control, self._wx_on_control_size)
        wx.EVT_MOVE(self.control, self._wx_on_control_move)
        wx.EVT_CHAR(self.control, self._wx_on_char)

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create_control(self, parent):
        # create a basic window control

        style = wx.DEFAULT_FRAME_STYLE \
                | wx.FRAME_NO_WINDOW_MENU \
                | wx.CLIP_CHILDREN

        control = wx.Frame(
            parent, -1, self.title, style=style, size=self.size,
            pos=self.position
        )

        control.SetBackgroundColour(SystemMetrics().dialog_background_color)

        return control

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_position(self):
        """ Property getter for position. """

        return self._position

    def _set_position(self, position):
        """ Property setter for position. """

        if self.control is not None:
            self.control.SetPosition(position)

        old = self._position
        self._position = position

        self.trait_property_changed('position', old, position)

    def _get_size(self):
        """ Property getter for size. """

        return self._size

    def _set_size(self, size):
        """ Property setter for size. """

        if self.control is not None:
            self.control.SetSize(size)

        old = self._size
        self._size = size

        self.trait_property_changed('size', old, size)

    def _title_changed(self, title):
        """ Static trait change handler. """

        if self.control is not None:
            self.control.SetTitle(title)

    #### wx event handlers ####################################################

    def _wx_on_activate(self, event):
        """ Called when the frame is being activated or deactivated. """

        if event.GetActive():
            self.activated = self
        else:
            self.deactivated = self

        event.Skip()

    def _wx_on_close(self, event):
        """ Called when the frame is being closed. """

        self.close()

    def _wx_on_control_move(self, event):
        """ Called when the window is resized. """

        # Get the real position and set the trait without performing
        # notification.

        # WXBUG - From the API documentation you would think that you could
        # call event.GetPosition directly, but that would be wrong.  The pixel
        # reported by that call is the pixel just below the window menu and
        # just right of the Windows-drawn border.
        self._position = event.GetEventObject().GetPositionTuple()

        event.Skip()

    def _wx_on_control_size(self, event):
        """ Called when the window is resized. """

        # Get the new size and set the shadow trait without performing
        # notification.
        wxsize = event.GetSize()

        self._size = (wxsize.GetWidth(), wxsize.GetHeight())

        event.Skip()

    def _wx_on_char(self, event):
        """ Called when a key is pressed when the tree has focus. """

        self.key_pressed = KeyPressedEvent(
            alt_down     = event.m_altDown == 1,
            control_down = event.m_controlDown == 1,
            shift_down   = event.m_shiftDown == 1,
            key_code     = event.m_keyCode,
            event        = event
        )

        event.Skip()

#### EOF ######################################################################



""" The default base class for extension providers. """


# Enthought library imports.
from traits.api import Event, HasTraits, provides

# Local imports.
from .extension_point_changed_event import ExtensionPointChangedEvent
from .i_extension_provider import IExtensionProvider


@provides(IExtensionProvider)
class ExtensionProvider(HasTraits):
    """ The default base class for extension providers. """

    #### 'IExtensionProvider' interface #######################################

    # The event fired when one of the provider's extension points has been
    # changed (where 'changed' means that the provider has added or removed
    # contributions to or from an extension point).
    extension_point_changed = Event(ExtensionPointChangedEvent)

    def get_extension_points(self):
        """ Return the extension points offered by the provider. """

        return []

    def get_extensions(self, extension_point_id):
        """ Return the provider's extensions to an extension point. """

        return []

    ##### Protected 'ExtensionProvider' interface #############################

    def _fire_extension_point_changed(
        self, extension_point_id, added, removed, index
    ):
        """ Fire an extension point changed event. """

        self.extension_point_changed = ExtensionPointChangedEvent(
            extension_point_id = extension_point_id,
            added              = added,
            removed            = removed,
            index              = index
        )

        return

#### EOF ######################################################################



# Run tests using nose:
#   $ nosetests -v
#
# The tests that are not compatible with the selected toolkit will be skipped
# (resulting in 'S' characters in the report). To test different backend,
# you should set the backend using the environment variable ETS_TOOLKIT:
#
#   $ ETS_TOOLKIT='qt4' nosetests -v
#   $ ETS_TOOLKIT='wx' nosetests -v
#   $ ETS_TOOLKIT='null' nosetests -v



from __future__ import absolute_import

import time

from traits.testing.unittest_tools import unittest

from ..action_event import ActionEvent


class TestActionEvent(unittest.TestCase):

    def test_init(self):
        t0 = time.time()
        event = ActionEvent()
        t1 = time.time()
        self.assertGreaterEqual(event.when, t0)
        self.assertLessEqual(event.when, t1)



from __future__ import (absolute_import, division, print_function)


# Follows the 4 aspects of service metadata

class IServiceIdentificationMetadata:
    """OO-interface to service identification metadata.
    """

    type = property("""Service name (string): "WMS", "WFS", "WCS", or "SOS".""")
    version = property("""Version of service protocol (string).""")
    title = property("""Human-targeted title of service (string).""")
    abstract = property("""Text describing the service (string).""")
    keywords = property("""Keyword list (list).""")
    accessconstraints = property("""Explanation of access constraints associated with service (string).""")
    fees = property("""Explanation of fees associated with service (string).""")
    

class IServiceProviderMetadata:
    """OO-interface to service provider metadata.
    """

    name = property("""Provider's name (string).""")
    url = property("""URL for provider's web site (string).""")
    contact = property("""How to contact the service provider (string).""")


class IServiceOperations:
    """OO-interface to service operations metadata.
    """

    operations = property("""List of operation descriptors (list). These must implement IOperationMetadata (below).""")
    exceptions = property("""List of exception formats (list).""")


class IServiceContents:
    """OO-interface to service contents metadata.
    """

    contents = property("""List of content descriptors (list). These must implement IServiceContent (below).""")


# IServiceMetadata aggregates the 4 aspects above

class IServiceMetadata(IServiceOperations, IServiceContents):
    """OWS Metadata.

    operations and contents properties are inherited.
    """

    identification = property("""Object that implements IServiceIdentificationMetadata.""")
    provider = property("""Object that implements IServiceProviderMetadata.""")


# A Service has an online resource URL as well as metadata collections

class IService(IServiceMetadata):
    """The OGC Web Service interface.
    """

    url = property("""Online resource URL (string)""")


# 3 specific service types are described below: WMS, WFS, and WCS

class IWebMapService(IService):
    """Abstraction for an OGC Web Map Service (WMS).
    """

    def getcapabilities():
        """Make a request to the WMS, returns an XML document wrapped in a 
        Python file object."""

    def getmap(**kw):
        """Make a request to the WMS, returns an image wrapped in a Python
        file object."""

    def getfeatureinfo(**kw):
        """Make a request to the WMS, returns data."""


class IWebFeatureService(IService):
    """Abstraction for an OGC Web Feature Service (WFS).
    """

    def getcapabilities():
        """Make a request to the WFS, returns an XML document wrapped in a 
        Python file object."""

    def getfeature(**kw):
        """Make a request to the WFS, returns an XML document wrapped in a
        Python file object."""

    def describefeaturetype(**kw):
        """Make a request to the WFS, returns data."""


class IWebCoverageService(IService):
    # TODO
    pass

class ISensorObservationService(IService):
    """Abstraction for an OGC Sensor Observation Service (SOS).
    """

    def getcapabilities():
        """Make a request to the SOS, returns an XML document wrapped in a 
        Python file object."""

    def describesensor():
        """Make a request to the SOS, returns an XML document wrapped in a 
        Python file object."""

    def getobservation():
        """Make a request to the SOS, returns an XML document wrapped in a 
        Python file object."""

# Second level metadata interfaces

class IOperationMetadata:
    """OO-interface to operation metadata.
    """

    name = property("""Operation name (string): GetCapabilities", for example.""")
    formatOptions = property("""List of content types (list).""")
    methods = property("""Mapping of method descriptors, keyed to HTTP verbs. Items must implement IMethodMetadata (below).""")


class IMethodMetadata:
    """OO-interface to method metadata.
    """

    url = property("""Method endpoint URL (string).""")
    # TODO: constraint

class IContentMetadata:
    """OO-interface to content metadata.
    """

    id = property("""Unique identifier (string).""")
    title = property("""Human-targeted title (string).""")
    boundingBox = property("""Four bounding values and a coordinate reference system identifier (tuple).""")
    boundingBoxWGS84 = property("""Four bounding values in WGS coordinates.""")
    crsOptions = property("""List of available coordinate/spatial reference systems (list).""")
    styles = property("""List of style dicts (list).""")
    timepositions=property("""List of times for which data is available""")
    defaulttimeposition=property("""Default time position""")

class iSensorObservationServiceContentMetadata(IContentMetadata):
    """Extension class for SOS specifics"""
    pass

# XXX: needed?

class IContactMetadata:
    """OO-interface to OWS metadata.

    Properties
    ----------
    name : string
    organization : string
    address : string
    city : string
    region : string
    postcode : string
    country : string
    email : string
    hoursofservice: string
    role: string
    """



#-------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   02/29/2008
#
#-------------------------------------------------------------------------

"""  Defines adapter interfaces for use with the ListStrEditor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (Any, Bool, Color, Enum, Event, HasPrivateTraits, Int,
                        Interface, List, Str, on_trait_change, provides)

#-------------------------------------------------------------------------
#  'IListStrAdapter' interface:
#-------------------------------------------------------------------------


class IListStrAdapter(Interface):

    # The index of the current item being adapted:
    index = Int

    # Current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    # Does the adapter know how to handler the current *item* or not:
    accepts = Bool

    # Does the value of *accepts* depend only upon the type of *item*?
    is_cacheable = Bool

#-------------------------------------------------------------------------
#  'AnIListStrAdapter' class:
#-------------------------------------------------------------------------


@provides(IListStrAdapter)
class AnIListStrAdapter(HasPrivateTraits):

    #-- Implementation of the IListStrAdapter Interface ----------------------

    # The index of the current item being adapted:
    index = Int

    # Current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    # Does the adapter know how to handler the current *item* or not:
    accepts = Bool(True)

    # Does the value of *accepts* depend only upon the type of *item*?
    is_cacheable = Bool(True)

#-------------------------------------------------------------------------
#  'ListStrAdapter' class:
#-------------------------------------------------------------------------


class ListStrAdapter(HasPrivateTraits):
    """ The base class for adapting list items to values that can be edited
        by a ListStrEditor.
    """

    #-- Trait Definitions ----------------------------------------------------

    # Specifies the default value for a new list item:
    default_value = Any('')

    # Specifies the default text for a new list item:
    default_text = Str

    # The default text color for list items (even, odd, any rows):
    even_text_color = Color(None, update=True)
    odd_text_color = Color(None, update=True)
    text_color = Color(None, update=True)

    # The default background color for list items (even, odd, any rows):
    even_bg_color = Color(None, update=True)
    odd_bg_color = Color(None, update=True)
    bg_color = Color(None, update=True)

    # The name of the default image to use for list items:
    image = Str(None, update=True)

    # Can the text value of each list item be edited:
    can_edit = Bool(True)

    # Specifies where a dropped item should be placed in the list relative to
    # the item it is dropped on:
    dropped = Enum('after', 'before')

    # The index of the current item being adapter:
    index = Int

    # The current item being adapted:
    item = Any

    # The current value (if any):
    value = Any

    # List of optional delegated adapters:
    adapters = List(IListStrAdapter, update=True)

    #-- Private Trait Definitions --------------------------------------------

    # Cache of attribute handlers:
    cache = Any({})

    # Event fired when the cache is flushed:
    cache_flushed = Event(update=True)

    #-- Adapter methods that are sensitive to item type ----------------------

    def get_can_edit(self, object, trait, index):
        """ Returns whether the user can edit a specified *object.trait[index]*
            list item. A True result indicates the value can be edited, while
            a False result indicates that it cannot be edited.
        """
        return self._result_for('get_can_edit', object, trait, index)

    def get_drag(self, object, trait, index):
        """ Returns the 'drag' value for a specified *object.trait[index]*
            list item. A result of *None* means that the item cannot be dragged.
        """
        return self._result_for('get_drag', object, trait, index)

    def get_can_drop(self, object, trait, index, value):
        """ Returns whether the specified *value* can be dropped on the
            specified *object.trait[index]* list item. A value of **True** means
            the *value* can be dropped; and a value of **False** indicates that
            it cannot be dropped.
        """
        return self._result_for('get_can_drop', object, trait, index, value)

    def get_dropped(self, object, trait, index, value):
        """ Returns how to handle a specified *value* being dropped on a
            specified *object.trait[index]* list item. The possible return
            values are:

            'before'
                Insert the specified *value* before the dropped on item.
            'after'
                Insert the specified *value* after the dropped on item.
        """
        return self._result_for('get_dropped', object, trait, index, value)

    def get_text_color(self, object, trait, index):
        """ Returns the text color for a specified *object.trait[index]* list
            item. A result of None means use the default list item text color.
        """
        return self._result_for('get_text_color', object, trait, index)

    def get_bg_color(self, object, trait, index):
        """ Returns the background color for a specified *object.trait[index]*
            list item. A result of None means use the default list item
            background color.
        """
        return self._result_for('get_bg_color', object, trait, index)

    def get_image(self, object, trait, index):
        """ Returns the name of the image to use for a specified
            *object.trait[index]* list item. A result of None means no image
            should be used. Otherwise, the result should either be the name of
            the image, or an ImageResource item specifying the image to use.
        """
        return self._result_for('get_image', object, trait, index)

    def get_item(self, object, trait, index):
        """ Returns the value of the *object.trait[index]* list item.
        """
        return self._result_for('get_item', object, trait, index)

    def get_text(self, object, trait, index):
        """ Returns the text to display for a specified *object.trait[index]*
            list item.
        """
        return self._result_for('get_text', object, trait, index)

    #-- Adapter methods that are not sensitive to item type ------------------

    def len(self, object, trait):
        """ Returns the number of items in the specified *object.trait* list.
        """
        # Sometimes, during shutdown, the object has been set to None.
        if object is None:
            return 0
        else:
            return len(getattr(object, trait))

    def get_default_value(self, object, trait):
        """ Returns a new default value for the specified *object.trait* list.
        """
        return self.default_value

    def get_default_text(self, object, trait):
        """ Returns the default text for the specified *object.trait* list.
        """
        return self.default_text

    def get_default_image(self, object, trait):
        """ Returns the default image for the specified *object.trait* list.
        """
        return self.image

    def get_default_bg_color(self, object, trait):
        """ Returns the default background color for the specified
            *object.trait* list.
        """
        return self._get_bg_color()

    def get_default_text_color(self, object, trait):
        """ Returns the default text color for the specified *object.trait*
            list.
        """
        return self._get_text_color()

    def set_text(self, object, trait, index, text):
        """ Sets the text for a specified *object.trait[index]* list item to
            *text*.
        """
        getattr(object, trait)[index] = text

    def delete(self, object, trait, index):
        """ Deletes the specified *object.trait[index]* list item.
        """
        del getattr(object, trait)[index]

    def insert(self, object, trait, index, value):
        """ Inserts a new value at the specified *object.trait[index]* list
            index.
        """
        getattr(object, trait)[index: index] = [value]

    #-- Private Adapter Implementation Methods -------------------------------

    def _get_can_edit(self):
        return self.can_edit

    def _get_drag(self):
        return unicode(self.item)

    def _get_can_drop(self):
        return isinstance(self.value, basestring)

    def _get_dropped(self):
        return self.dropped

    def _get_text_color(self):
        if (self.index % 2) == 0:
            return self.even_text_color_ or self.text_color_

        return self.odd_text_color or self.text_color_

    def _get_bg_color(self):
        if (self.index % 2) == 0:
            return self.even_bg_color_ or self.bg_color_

        return self.odd_bg_color or self.bg_color_

    def _get_image(self):
        return self.image

    def _get_item(self):
        return self.item

    def _get_text(self):
        return unicode(self.item)

    #-- Private Methods ------------------------------------------------------

    def _result_for(self, name, object, trait, index, value=None):
        """ Returns/Sets the value of the specified *name* attribute for the
            specified *object.trait[index]* list item.
        """
        self.index = index
        self.value = value
        items = getattr(object, trait)
        if index >= len(items):
            self.item = item = None
        else:
            self.item = item = items[index]

        item_class = item.__class__
        key = '%s:%s' % (item_class.__name__, name)
        handler = self.cache.get(key)
        if handler is not None:
            return handler()

        trait_name = name[4:]

        for adapter in self.adapters:
            adapter.index = index
            adapter.item = item
            adapter.value = value
            if adapter.accepts and (adapter.trait(trait_name) is not None):
                handler = lambda: getattr(
                    adapter.set(
                        index=self.index,
                        item=self.item,
                        value=self.value),
                    trait_name)

                if adapter.is_cacheable:
                    break

                return handler()
        else:
            for klass in item_class.__mro__:
                cname = '%s_%s' % (klass.__name__, trait_name)
                if self.trait(cname) is not None:
                    handler = lambda: getattr(self, cname)
                    break
            else:
                handler = getattr(self, '_' + name)

        self.cache[key] = handler
        return handler()

    @on_trait_change('adapters.+update')
    def _flush_cache(self):
        """ Flushes the cache when any trait on any adapter changes.
        """
        self.cache = {}
        self.cache_flushed = True
